# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.topology import _assert
from tobiko.openstack.topology import _exception
from tobiko.openstack.topology import _neutron
from tobiko.openstack.topology import _topology

assert_reachable_nodes = _assert.assert_reachable_nodes
assert_unreachable_nodes = _assert.assert_unreachable_nodes

NoSuchOpenStackTopologyNodeGroup = _exception.NoSuchOpenStackTopologyNodeGroup
NoSuchOpenStackTopologyNode = _exception.NoSuchOpenStackTopologyNode

UnknowOpenStackContainerNameError = _topology.UnknowOpenStackContainerNameError
UnknowOpenStackServiceNameError = _topology.UnknowOpenStackServiceNameError
get_agent_service_name = _topology.get_agent_service_name
get_agent_container_name = _topology.get_agent_container_name
get_log_file_digger = _topology.get_log_file_digger
get_openstack_topology = _topology.get_openstack_topology
get_openstack_node = _topology.get_openstack_node
get_openstack_version = _topology.get_openstack_version
skip_unless_osp_version = _topology.skip_unless_osp_version
check_systemd_monitors_agent = _topology.check_systemd_monitors_agent
find_openstack_node = _topology.find_openstack_node
get_default_openstack_topology_class = (
    _topology.get_default_openstack_topology_class)
list_openstack_nodes = _topology.list_openstack_nodes
list_openstack_node_groups = _topology.list_openstack_node_groups
OpenStackTopology = _topology.OpenStackTopology
OpenStackTopologyNode = _topology.OpenStackTopologyNode
set_default_openstack_topology_class = (
    _topology.set_default_openstack_topology_class)
verify_osp_version = _topology.verify_osp_version

NeutronNovaResponse = _neutron.NeutronNovaResponse
NeutronNovaResponseReader = _neutron.NeutronNovaResponseReader
read_neutron_nova_responses = _neutron.read_neutron_nova_responses



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.octavia import _client
from tobiko.openstack.octavia import _waiters
from tobiko.openstack.octavia import _constants
from tobiko.openstack.octavia import _validators
from tobiko.openstack.octavia import _exceptions


OCTAVIA_CLIENT_CLASSSES = _client.OCTAVIA_CLIENT_CLASSSES
get_octavia_client = _client.get_octavia_client
octavia_client = _client.octavia_client
OctaviaClientFixture = _client.OctaviaClientFixture
get_loadbalancer = _client.get_loadbalancer
get_member = _client.get_member

# Waiters
wait_for_status = _waiters.wait_for_status

# Validators
check_members_balanced = _validators.check_members_balanced

# Exceptions
RequestException = _exceptions.RequestException
TimeoutException = _exceptions.TimeoutException

# Constants
PROVISIONING_STATUS = _constants.PROVISIONING_STATUS
ACTIVE = _constants.ACTIVE
ERROR = _constants.ERROR



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import sys

from tobiko.common import _asserts
from tobiko.common import _cached
from tobiko.common import _config
from tobiko.common import _detail
from tobiko.common import _exception
from tobiko.common import _fixture
from tobiko.common import _logging
from tobiko.common.managers import loader as loader_manager
from tobiko.common import _operation
from tobiko.common import _os
from tobiko.common import _proxy
from tobiko.common import _retry
from tobiko.common import _select
from tobiko.common import _skip
from tobiko.common import _testcase
from tobiko.common import _time
from tobiko.common import _utils


TOBIKO_PACKAGE_DIR = os.path.dirname(os.path.realpath(__file__))

# Ensure any tobiko package subdir is in sys.path
for path_dir in list(sys.path):
    path_dir = os.path.realpath(path_dir)
    if path_dir.startswith(TOBIKO_PACKAGE_DIR):
        sys.path.remove(path_dir)


details_content = _detail.details_content

FailureException = _asserts.FailureException
fail = _asserts.fail

cached = _cached.cached
CachedProperty = _cached.CachedProperty

tobiko_config = _config.tobiko_config
tobiko_config_dir = _config.tobiko_config_dir
tobiko_config_path = _config.tobiko_config_path

TobikoException = _exception.TobikoException
check_valid_type = _exception.check_valid_type
exc_info = _exception.exc_info
handle_multiple_exceptions = _exception.handle_multiple_exceptions
list_exc_infos = _exception.list_exc_infos

is_fixture = _fixture.is_fixture
get_fixture = _fixture.get_fixture
fixture_property = _fixture.fixture_property
required_fixture = _fixture.required_fixture
required_setup_fixture = _fixture.required_setup_fixture
get_fixture_name = _fixture.get_fixture_name
get_fixture_class = _fixture.get_fixture_class
get_fixture_dir = _fixture.get_fixture_dir
get_object_name = _fixture.get_object_name
remove_fixture = _fixture.remove_fixture
reset_fixture = _fixture.reset_fixture
setup_fixture = _fixture.setup_fixture
cleanup_fixture = _fixture.cleanup_fixture
list_required_fixtures = _fixture.list_required_fixtures
SharedFixture = _fixture.SharedFixture
FixtureManager = _fixture.FixtureManager

CaptureLogFixture = _logging.CaptureLogFixture

load_object = loader_manager.load_object
load_module = loader_manager.load_module

makedirs = _os.makedirs
open_output_file = _os.open_output_file

runs_operation = _operation.runs_operation
before_operation = _operation.before_operation
after_operation = _operation.after_operation
with_operation = _operation.with_operation
RunsOperations = _operation.RunsOperations
Operation = _operation.Operation
get_operation = _operation.get_operation
get_operation_name = _operation.get_operation_name
operation_config = _operation.operation_config

call_proxy = _proxy.call_proxy
call_proxy_class = _proxy.call_proxy_class
list_protocols = _proxy.list_protocols
protocol = _proxy.protocol
CallHandler = _proxy.CallHandler

retry = _retry.retry
retry_attempt = _retry.retry_attempt
retry_on_exception = _retry.retry_on_exception
retry_test_case = _retry.retry_test_case
Retry = _retry.Retry
RetryAttempt = _retry.RetryAttempt
RetryCountLimitError = _retry.RetryCountLimitError
RetryLimitError = _retry.RetryLimitError
RetryTimeLimitError = _retry.RetryTimeLimitError

Selection = _select.Selection
select = _select.select
ObjectNotFound = _select.ObjectNotFound
MultipleObjectsFound = _select.MultipleObjectsFound

SkipException = _skip.SkipException
skip_if = _skip.skip_if
skip_test = _skip.skip_test
skip_unless = _skip.skip_unless
skip = _skip.skip

assert_test_case_was_skipped = _testcase.assert_test_case_was_skipped
get_test_case = _testcase.get_test_case
pop_test_case = _testcase.pop_test_case
push_test_case = _testcase.push_test_case
run_test = _testcase.run_test
TestCasesManager = _testcase.TestCasesManager

min_seconds = _time.min_seconds
max_seconds = _time.max_seconds
Seconds = _time.Seconds
SecondsValueError = _time.SecondsValueError
sleep = _time.sleep
time = _time.time
to_seconds = _time.to_seconds
to_seconds_float = _time.to_seconds_float
true_seconds = _time.true_seconds

get_short_hostname = _utils.get_short_hostname

from tobiko import config  # noqa
config.init_config()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections

import tobiko
from tobiko.openstack.nova import _client


class HypervisorListFixture(tobiko.SharedFixture):

    client = None
    hypervisors = None

    def setup_fixture(self):
        self.setup_client()
        self.get_hypervisors()

    def setup_client(self):
        self.client = _client.get_nova_client()

    def get_hypervisors(self):
        self.hypervisors = _client.list_hypervisors(client=self.client)


def get_hypervisors(**params):
    hypervisors = tobiko.setup_fixture(HypervisorListFixture).hypervisors
    return tobiko.select(hypervisors).with_attributes(**params)


def missing_hypervisors(count=1, **params):
    agents = get_hypervisors(**params)
    return max(0, count - len(agents))


def skip_if_missing_hypervisors(count=1, **params):
    message = "missing {return_value!r} hypervisor(s)"
    if params:
        message += " with {!s}".format(
            ', '.join("{!s}={!r}".format(k, v) for k, v in params.items()))
    return tobiko.skip_if(message, missing_hypervisors, count=count,
                          **params)


def get_same_host_hypervisors(servers, hypervisor):
    host_hypervisors = get_servers_hypervisors(servers)
    same_host_server_ids = host_hypervisors.pop(hypervisor, None)
    if same_host_server_ids:
        return {hypervisor: same_host_server_ids}
    else:
        return {}


def get_different_host_hypervisors(servers, hypervisor):
    host_hypervisors = get_servers_hypervisors(servers)
    host_hypervisors.pop(hypervisor, None)
    return host_hypervisors


def get_servers_hypervisors(servers, client=None):
    hypervisors = collections.defaultdict(list)
    for server in (servers or list()):
        client = _client.nova_client(client)
        if isinstance(server, str):
            server_id = server
            server = _client.get_server(server_id, client=client)
        else:
            server_id = server.id
        hypervisor = get_server_hypervisor(server)
        hypervisors[hypervisor].append(server_id)
    return hypervisors


def get_server_hypervisor(server, client=None):
    if isinstance(server, str):
        server = _client.get_server(server, client=client)
    return getattr(server, 'OS-EXT-SRV-ATTR:host')



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from datetime import datetime
import os
import subprocess

from oslo_log import log
from py.xml import html  # pylint: disable=no-name-in-module,import-error
import pytest

import tobiko

LOG = log.getLogger(__name__)

TOX_REPORT_NAME = os.environ.get('TOX_REPORT_NAME', "tobiko_results")


@pytest.hookimpl
def pytest_configure(config):
    configure_metadata(config)
    configure_caplog(config)
    configure_timeout(config)
    configure_junitxml(config)


def configure_metadata(config):
    # pylint: disable=protected-access
    from tobiko import version
    config._metadata["Tobiko Version"] = version.release
    git_commit = subprocess.check_output(
        ['git', 'log', '-n', '1'],
        universal_newlines=True).replace('\n', '<br>')
    config._metadata["Tobiko Git Commit"] = git_commit


def configure_caplog(config):
    tobiko_config = tobiko.tobiko_config()

    if tobiko_config.logging.capture_log is True:
        if tobiko_config.debug:
            default = 'DEBUG'
        else:
            default = 'INFO'
    else:
        default = 'FATAL'
    for key in ['log_level',
                'log_file_level',
                'log_cli_level']:
        set_default_inicfg(config, key, default)

    default = tobiko_config.logging_default_format_string
    if default:
        # instance and color are not supported by pytest
        default = default.replace('%(instance)s', '')
        default = default.replace('%(color)s', '')
        if default:
            for key in ['log_format',
                        'log_file_format',
                        'log_cli_format']:
                set_default_inicfg(config, key, default)

    default = tobiko_config.log_date_format
    if default:
        for key in ['log_date_format',
                    'log_file_date_format',
                    'log_cli_date_format']:
            set_default_inicfg(config, key, default)


def configure_junitxml(config):
    config.inicfg['junit_suite_name'] = TOX_REPORT_NAME


def set_default_inicfg(config, key, default):
    value = config.inicfg.setdefault(key, default)
    if value != default:
        LOG.debug(f"Set default inicfg: {key} = {value}")
    else:
        LOG.debug(f"Keep existing inicfg: {key} = {value}")


class TestRunnerTimeoutManager(tobiko.SharedFixture):
    timeout: tobiko.Seconds = None
    deadline: tobiko.Seconds = None

    def setup_fixture(self):
        tobiko_config = tobiko.tobiko_config()
        self.timeout = tobiko_config.testcase.test_runner_timeout
        if self.timeout is None:
            LOG.info('Test runner timeout is disabled')
        else:
            LOG.info('Test runner timeout is enabled: '
                     f'timeout is {self.timeout} seconds')
            self.deadline = tobiko.time() + self.timeout

    @classmethod
    def check_test_runner_timeout(cls):
        self = tobiko.setup_fixture(cls)
        if self.deadline is not None:
            time_left = self.deadline - tobiko.time()
            if time_left <= 0.:
                pytest.skip(
                    f"Test runner execution timed out after {self.timeout} "
                    f"seconds",
                    allow_module_level=True)
            else:
                LOG.debug('Test runner timeout is enabled: '
                          f'{time_left} seconds left')


def check_test_runner_timeout():
    TestRunnerTimeoutManager.check_test_runner_timeout()


def configure_timeout(config):
    tobiko_config = tobiko.tobiko_config()
    default = tobiko_config.testcase.timeout
    if default is not None and default > 0.:
        set_default_inicfg(config, 'timeout', default)


def pytest_html_results_table_header(cells):
    cells.insert(2, html.th("Description"))
    cells.insert(1, html.th("Time", class_="sortable time", col="time"))
    cells.pop()


def pytest_html_results_table_row(report, cells):
    cells.insert(2, html.td(getattr(report, 'description', '')))
    cells.insert(1, html.td(datetime.utcnow(), class_="col-time"))
    cells.pop()


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    # pylint: disable=unused-argument
    outcome = yield
    report = outcome.get_result()
    report.description = getattr(item.function, '__doc__', '')


def pytest_html_report_title(report):
    report.title = f"Tobiko test results ({TOX_REPORT_NAME})"


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_call(item):
    # pylint: disable=protected-access
    check_test_runner_timeout()
    tobiko.push_test_case(item._testcase)
    try:
        yield
    finally:
        tobiko.pop_test_case()



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import testtools


FailureException = testtools.TestCase.failureException


def fail(msg: str, *args, **kwargs) -> typing.NoReturn:
    """Fail immediately current test case execution, with the given message.

    Unconditionally raises a tobiko.FailureException as in below equivalent
    code:

        raise FailureException(msg.format(*args, **kwargs))

    :param msg: string message used to create FailureException
    :param *args: positional arguments to be passed to str.format method
    :param **kwargs: key-word arguments to be passed to str.format method
    :returns: It never returns
    :raises FailureException:
    """
    if args or kwargs:
        msg = msg.format(*args, **kwargs)
    raise FailureException(msg)



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import re
import time
import typing

from oslo_log import log

import tobiko
from tobiko.shell.sh import _cmdline
from tobiko.shell.sh import _command
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _hostname
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class PsError(tobiko.TobikoException):
    message = "Unable to list processes from host: {error}"


class PsWaitTimeout(PsError):
    message = ("Process(es) still running on host {hostname!r} after "
               "{timeout} seconds:\n{processes!s}")


IS_KERNEL_RE = re.compile('^\\[.*\\]$')


_NOT_FOUND = object()


class PsProcessBase:
    command: str
    pid: int
    ssh_client: ssh.SSHClientType

    @property
    def is_kernel(self) -> bool:
        return IS_KERNEL_RE.match(self.command) is not None

    @property
    def command_line(self) -> typing.Optional[_command.ShellCommand]:
        try:
            return _cmdline.get_command_line(command=self.command,
                                             pid=self.pid,
                                             ssh_client=self.ssh_client,
                                             _cache_id=id(self))
        except _cmdline.GetCommandLineError as ex:
            LOG.error(str(ex))
            return None


class PsProcessTuple(typing.NamedTuple):
    """Process listed by ps command
    """
    command: str
    pid: int
    ssh_client: ssh.SSHClientType


class PsProcess(PsProcessTuple, PsProcessBase):
    pass


P = typing.TypeVar('P', bound=PsProcessBase)


def select_processes(
        processes: typing.Iterable[PsProcessBase],
        command: str = None,
        pid: int = None,
        is_kernel: typing.Optional[bool] = False,
        command_line: _command.ShellCommandType = None) \
        -> tobiko.Selection[P]:
    selection = tobiko.Selection[P](processes)

    if selection and pid is not None:
        # filter files by PID
        selection = selection.with_attributes(pid=pid)

    if selection and command_line is not None:
        if command is None:
            command = _command.shell_command(command_line)[0]

    if selection and command is not None:
        # filter processes by command
        pattern = re.compile(command)
        selection = selection.select(
            lambda process: bool(pattern.match(str(process.command))))

    if selection and is_kernel is not None:
        # filter kernel processes
        selection = selection.with_attributes(is_kernel=bool(is_kernel))

    if selection and command_line is not None:
        pattern = re.compile(str(command_line))
        selection = selection.select(
            lambda process: bool(pattern.match(str(process.command_line))))

    return selection


def list_kernel_processes(**list_params):
    return list_processes(is_kernel=True, **list_params)


def list_all_processes(**list_params):
    return list_processes(is_kernel=None, **list_params)


def list_processes(
        pid: int = None,
        command: str = None,
        is_kernel: typing.Optional[bool] = False,
        ssh_client: ssh.SSHClientType = None,
        command_line: _command.ShellCommandType = None,
        **execute_params) -> tobiko.Selection[PsProcess]:
    """Returns list of running process

    """
    result = _execute.execute('ps -A', expect_exit_status=None,
                              ssh_client=ssh_client, **execute_params)
    output = result.stdout and result.stdout.strip()
    if result.exit_status or not output:
        raise PsError(error=result.stderr)

    # Extract a list of PsProcess instances from table body
    processes = tobiko.Selection[PsProcess]()
    for process_data in parse_table(lines=output.splitlines(),
                                    schema=PS_TABLE_SCHEMA):
        processes.append(PsProcess(ssh_client=ssh_client, **process_data))

    return select_processes(processes,
                            pid=pid,
                            command=command,
                            is_kernel=is_kernel,
                            command_line=command_line)


def wait_for_processes(timeout=float('inf'), sleep_interval=5.,
                       ssh_client=None, **list_params):
    start_time = time.time()
    time_left = timeout
    while True:
        processes = list_processes(timeout=time_left,
                                   ssh_client=ssh_client,
                                   **list_params)
        if not processes:
            break

        time_left = timeout - (time.time() - start_time)
        if time_left < sleep_interval:
            hostname = _hostname.get_hostname(ssh_client=ssh_client)
            process_lines = [
                '    {pid} {command}'.format(pid=process.pid,
                                             command=process.command)
                for process in processes]
            raise PsWaitTimeout(timeout=timeout, hostname=hostname,
                                processes='\n'.join(process_lines))

        time.sleep(sleep_interval)


def parse_pid(value):
    return 'pid', int(value)


def parse_command(value):
    return 'command', str(value)


PS_TABLE_SCHEMA = {
    'pid': parse_pid,
    'cmd': parse_command,
    'command': parse_command,
}


def parse_table(lines, schema, header_line=None):
    lines = iter(lines)
    while not header_line:
        header_line = next(lines)

    getters = []
    column_names = header_line.strip().lower().split()
    for position, name in enumerate(column_names):
        getter = schema.get(name)
        if getter:
            getters.append((position, getter))

    for line in lines:
        row = line.strip().split()
        if row:
            yield dict(getter(row[position])
                       for position, getter in getters)



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import typing  # noqa
import time

# import testtools

import tobiko
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import nova
from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _nova


class TestServerCreationStack(_cirros.CirrosServerStackFixture):
    """Nova instance intended to be used for testing server creation"""


def test_server_creation(stack=TestServerCreationStack):
    """Test Nova server creation
    """
    return test_servers_creation(stack=stack,
                                 number_of_servers=0).first


class TestEvacuableServerCreationStack(_cirros.EvacuableServerStackFixture):
    """Nova instance intended to be used for testing server creation"""


def test_evacuable_server_creation():
    """Test evacuable Nova server creation
    """
    return test_server_creation(stack=TestEvacuableServerCreationStack)


def test_server_creation_and_shutoff(stack=TestServerCreationStack):
    result = test_server_creation(stack=stack)
    nova.shutoff_server(result.server_details)
    return result


def test_servers_creation(stack=TestServerCreationStack,
                          number_of_servers=2) -> \
        tobiko.Selection[_nova.ServerStackFixture]:

    initial_servers_ids = {server.id for server in nova.list_servers()}
    pid = os.getpid()
    fixture_obj = tobiko.get_fixture_class(stack)

    # Get list of server stack instances
    fixtures: tobiko.Selection[_nova.ServerStackFixture] = tobiko.select(
        tobiko.get_fixture(fixture_obj, fixture_id=f'{pid}-{i}')
        for i in range(number_of_servers or 1))

    test_case = tobiko.get_test_case()

    # Check fixtures types
    for fixture in fixtures:
        test_case.assertIsInstance(fixture, _nova.ServerStackFixture)

    # Delete all servers stacks
    for fixture in fixtures:
        tobiko.cleanup_fixture(fixture)

    # Create all servers stacks
    for fixture in fixtures:
        test_case.useFixture(fixture)

    # Check every server ID is unique and new
    server_ids = {fixture.server_id for fixture in fixtures}
    test_case.assertEqual(number_of_servers or 1, len(server_ids))
    test_case.assertFalse(server_ids & initial_servers_ids)

    # sleep for 20 sec , ensure no race condition with ssh
    time.sleep(20)

    # Test SSH connectivity to floating IP address
    for fixture in fixtures:
        test_case.assertTrue(sh.get_hostname(ssh_client=fixture.ssh_client))

    # Test pinging to floating IP address
    ping.assert_reachable_hosts(fixture.floating_ip_address
                                for fixture in fixtures)
    return fixtures



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import socket
from unittest import mock

import paramiko

from tobiko.shell import sh
from tobiko.shell import ssh
from tobiko.tests import unit


class HostnameTest(unit.TobikoUnitTest):

    def mock_ssh_client(self,
                        stdout='mocked-hostname\n',
                        stderr='',
                        exit_status=0) \
            -> ssh.SSHClientFixture:
        channel_mock = mock.MagicMock(spec=paramiko.Channel,
                                      exit_status=exit_status)
        channel_mock.recv.side_effect = [bytes(stdout, 'utf-8'),
                                         EOFError,
                                         EOFError] * 10
        channel_mock.recv_stderr.side_effect = [bytes(stderr, 'utf-8'),
                                                EOFError,
                                                EOFError] * 10

        client_mock = mock.MagicMock(spec=ssh.SSHClientFixture)
        client_mock.connect().get_transport().open_session.return_value = \
            channel_mock
        return client_mock

    def test_get_hostname_with_no_ssh_client(self):
        hostname = sh.get_hostname(ssh_client=False)
        self.assertEqual(socket.gethostname(), hostname)

    def test_get_hostname_with_ssh_client(self):
        ssh_client = self.mock_ssh_client()
        hostname = sh.get_hostname(ssh_client=ssh_client)
        self.assertEqual('mocked-hostname', hostname)
        self.assertIs(hostname,
                      sh.get_hostname(ssh_client=ssh_client))

    def test_get_hostname_with_no_cached(self):
        ssh_client = self.mock_ssh_client()
        hostname = sh.get_hostname(ssh_client=ssh_client,
                                   cached=False)
        self.assertEqual('mocked-hostname', hostname)
        self.assertIsNot(hostname,
                         sh.get_hostname(ssh_client=ssh_client,
                                         cached=False))

    def test_get_hostname_with_ssh_proxy(self):
        ssh_client = self.mock_ssh_client()
        self.patch(ssh, 'ssh_client_fixture', return_value=ssh_client)
        hostname = sh.get_hostname(ssh_client=None)
        self.assertEqual('mocked-hostname', hostname)

    def test_get_hostname_with_ssh_client_no_output(self):
        ssh_client = self.mock_ssh_client(stdout='\n')
        ex = self.assertRaises(sh.HostNameError,
                               sh.get_hostname,
                               ssh_client=ssh_client)
        self.assertIn('Invalid result', str(ex))

    def test_get_hostname_with_ssh_client_and_failure(self):
        ssh_client = self.mock_ssh_client(exit_status=1,
                                          stderr='command not found')
        ex = self.assertRaises(sh.HostNameError,
                               sh.get_hostname,
                               ssh_client=ssh_client)
        self.assertIn('command not found', str(ex))



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import re
import typing

import tobiko
from tobiko.openstack.neutron import _client


# Agent binary names
DHCP_AGENT = 'neutron-dhcp-agent'
L3_AGENT = 'neutron-l3-agent'
METADATA_AGENT = 'neutron-metadata-agent'
OPENVSWITCH_AGENT = 'neutron-openvswitch-agent'
OVN_CONTROLLER = 'ovn-controller'
# NOTE(slaweq) name 'networking-ovn-metadata-agent' was used up to stable/train
# release when ovn driver was stadium project,
# Since stable/ussuri, when driver was included in the Neutron repo, it is
# called 'neutron-ovn-metadata-agent'
OVN_METADATA_AGENT = 'networking-ovn-metadata-agent'
NEUTRON_OVN_METADATA_AGENT = 'neutron-ovn-metadata-agent'


class AgentNotFoundOnHost(tobiko.TobikoException):
    message = "Agent {agent_type!s} not found on the host {host!s}"


NeutronAgentType = typing.Dict[str, typing.Any]


def list_agents(client=None, **params) \
        -> tobiko.Selection[NeutronAgentType]:
    agents = _client.neutron_client(client).list_agents(**params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.Selection[NeutronAgentType](agents)


def list_l3_agent_hosting_routers(router, client=None, **params):
    agents = _client.neutron_client(client).list_l3_agent_hosting_routers(
        router, **params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.select(agents)


def find_l3_agent_hosting_router(router, client=None, unique=False,
                                 **list_params):
    agents = list_l3_agent_hosting_routers(router=router, client=client,
                                           **list_params)
    if unique:
        return agents.unique
    else:
        return agents.first


def list_dhcp_agent_hosting_network(network, client=None, **params):
    agents = _client.neutron_client(client).list_dhcp_agent_hosting_networks(
        network, **params)
    if isinstance(agents, collections.Mapping):
        agents = agents['agents']
    return tobiko.select(agents)


class NetworkingAgentFixture(tobiko.SharedFixture):

    agents = None

    def setup_fixture(self):
        self.agents = list_agents()


def list_networking_agents(**attributes):
    return tobiko.setup_fixture(
        NetworkingAgentFixture).agents.with_items(**attributes)


def count_networking_agents(**params) -> int:
    return len(list_networking_agents(**params))


def missing_networking_agents(count=1, **params) -> int:
    actual_count = count_networking_agents(**params)
    return max(0, count - actual_count)


def has_networking_agents(**params) -> bool:
    return count_networking_agents(**params) > 0


def has_ovn() -> bool:
    return not has_ovs()


def has_ovs() -> bool:
    return has_networking_agents(binary=OPENVSWITCH_AGENT)


DecoratorType = typing.Callable[[typing.Union[typing.Callable, typing.Type]],
                                typing.Union[typing.Callable, typing.Type]]


AgentBinaryType = typing.Union[str, typing.Pattern[str]]


def skip_if_missing_networking_agents(
        binary: AgentBinaryType = None,
        count: int = 1,
        **params) \
        -> DecoratorType:
    if binary is not None:
        params['binary'] = binary
    message = "missing {return_value!r} agent(s)"
    if params:
        message += " with {!s}".format(
            ', '.join("{!s}={!r}".format(k, v) for k, v in params.items()))
    return tobiko.skip_if(message, missing_networking_agents, count=count,
                          **params)


def skip_unless_missing_networking_agents(
        binary: AgentBinaryType = None,
        count: int = 1,
        **params) \
        -> DecoratorType:
    if binary is not None:
        params['binary'] = binary
    message = "found {return_value!r} agent(s)"
    if params:
        message += " with {!s}".format(
            ', '.join("{!s}={!r}".format(k, v) for k, v in params.items()))
    return tobiko.skip_unless(message, missing_networking_agents, count=count,
                              **params)


def skip_if_is_old_ovn():
    """Skip the test if OVN is not configured"""
    binary = re.compile(f'({OPENVSWITCH_AGENT}|{OVN_CONTROLLER})')
    return skip_if_missing_networking_agents(binary)


def skip_unless_is_ovn():
    """Skip the test if OVN is not configured"""
    return skip_unless_missing_networking_agents(OPENVSWITCH_AGENT)


def skip_unless_is_ovs():
    """Skip the test if openvswitch agent does exist"""
    return skip_if_missing_networking_agents(OPENVSWITCH_AGENT)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import topology
from tobiko.shell import ping
from tobiko.shell import sh


@keystone.skip_unless_has_keystone_credentials()
class OpenStackTopologyTest(testtools.TestCase):

    topology = tobiko.required_setup_fixture(topology.OpenStackTopology)

    def test_get_openstack_topology(self):
        topology_class = type(self.topology)
        topo = topology.get_openstack_topology(topology_class=topology_class)
        self.assertIs(topo, self.topology)
        self.assertIsInstance(topo, topology.OpenStackTopology)

    def test_ping_node(self):
        for node in self.topology.nodes:
            ping.ping(node.public_ip, count=1, timeout=5.).assert_replied()

    def test_ssh_client(self):
        for node in self.topology.nodes:
            self.assertIsNotNone(node.ssh_client)
            hostname = sh.get_hostname(
                ssh_client=node.ssh_client).split('.')[0]
            self.assertEqual(node.name, hostname)

    def test_controller_group(self):
        nodes = list(self.topology.get_group('controller'))
        self.assertNotEqual([], nodes)
        for node in nodes:
            self.assertIn('controller', node.groups)

    def test_compute_group(self):
        nodes = list(self.topology.get_group('compute'))
        self.assertNotEqual([], nodes)
        for node in nodes:
            self.assertIn('compute', node.groups)
        hypervisors = {
            hypervisor.hypervisor_hostname.split('.', 1)[0].lower(): hypervisor
            for hypervisor in nova.list_hypervisors()}
        for name, hypervisor in hypervisors.items():
            node = self.topology.get_node(name)
            self.assertEqual(name, node.name)
            self.assertIn(node, nodes)

    def test_list_openstack_topology(self, group=None, hostnames=None):
        nodes = topology.list_openstack_nodes(
            topology=self.topology, group=group, hostnames=hostnames)
        self.assertTrue(set(nodes).issubset(set(self.topology.nodes)))
        for node in nodes:
            if group:
                self.assertIn(group, node.groups)
            if hostnames:
                hostnames = [node_name_from_hostname(h)
                             for h in hostnames]
                self.assertIn(node.name, hostnames)
        return nodes

    def test_list_openstack_topology_with_group(self):
        self.test_list_openstack_topology(group='compute')

    def test_list_openstack_topology_with_hostnames(self):
        expected_nodes = self.topology.nodes[0::2]
        actual_nodes = self.test_list_openstack_topology(
            hostnames=[node.name for node in expected_nodes])
        self.assertEqual(expected_nodes, actual_nodes)


def node_name_from_hostname(hostname):
    return hostname.split('.', 1)[0].lower()



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import os

from oslo_log import log

import tobiko
from tobiko.shell.sh import _command
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class GetCommandLineError(tobiko.TobikoException):
    message = "Unable to get process command line: {error}"


class GetCommandLineMismatch(GetCommandLineError):
    message = ("Command line of process ({pid}) doesn't match its command "
               "({command}): {command_line}")


@functools.lru_cache(typed=True)
def get_command_line(pid: int,
                     ssh_client: ssh.SSHClientType = None,
                     command: str = None,
                     _cache_id: int = None) \
        -> _command.ShellCommand:
    try:
        output = _execute.execute(f'cat /proc/{pid}/cmdline',
                                  ssh_client=ssh_client).stdout
    except _exception.ShellCommandFailed as ex:
        raise GetCommandLineError(error=ex.stderr) from ex

    command_line = _command.ShellCommand(output.strip().split('\0')[:-1])
    if not command_line:
        raise GetCommandLineError(error="command line is empty")

    if command is not None and os.path.basename(command_line[0]) != command:
        raise GetCommandLineMismatch(pid=pid, command=command,
                                     command_line=command_line)
    return command_line



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.shell.ssh import _config
from tobiko.shell.ssh import _client
from tobiko.shell.ssh import _command
from tobiko.shell.ssh import _forward
from tobiko.shell.ssh import _skip


SSHHostConfig = _config.SSHHostConfig
ssh_host_config = _config.ssh_host_config

SSHClientFixture = _client.SSHClientFixture
ssh_client = _client.ssh_client
ssh_command = _command.ssh_command
ssh_proxy_client = _client.ssh_proxy_client
SSHConnectFailure = _client.SSHConnectFailure
gather_ssh_connect_parameters = _client.gather_ssh_connect_parameters
SSHClientType = _client.SSHClientType
ssh_client_fixture = _client.ssh_client_fixture


reset_default_ssh_port_forward_manager = \
    _forward.reset_default_ssh_port_forward_manager
get_port_forward_url = _forward.get_forward_url
get_forward_port_address = _forward.get_forward_port_address
SSHTunnelForwarderFixture = _forward.SSHTunnelForwarderFixture
SSHTunnelForwarder = _forward.SSHTunnelForwarder

has_ssh_proxy_jump = _skip.has_ssh_proxy_jump
skip_unless_has_ssh_proxy_jump = _skip.skip_unless_has_ssh_proxy_jump



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


import sys


import tobiko


def import_lzma():
    try:
        import lzma
    except ImportError:
        from backports import lzma
    return lzma


def has_lzma():
    try:
        return import_lzma()
    except ImportError:
        return None


def open_file(filename, mode):
    try:
        lzma = import_lzma()
    except ImportError:
        tobiko.skip_test(
            "Package lzma or backports.lzma is required to decompress "
            f"{filename!r} (mode={mode!r}) XZ image file "
            f"({sys.version!r}).")

    return lzma.LZMAFile(filename=filename, mode=mode)



# Copyright (c) 2020 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

from tobiko.tests.faults.containers import container_ops


LOG = log.getLogger(__name__)


@container_ops.skip_unless_has_docker
class LogFilesTest(testtools.TestCase):

    def test_neutron_logs_exist(self):
        groups = ['controller', 'compute', 'networker']
        neutron_nodes = container_ops.get_nodes_for_groups(groups)
        for node in neutron_nodes:
            # set is used to remove duplicated containers
            containers = set(container_ops.get_node_neutron_containers(node) +
                             container_ops.get_node_ovn_containers(node))
            for container in containers:
                logfiles = (container_ops.
                            get_container_logfiles(node, container))
                if not logfiles:
                    LOG.warning(f'No logfiles have been found in {container} '
                                f'container of {node.name} node')
                    continue
                # logdir is obtained differently for pcs resources
                pcs_logdir = (container_ops.
                              get_node_logdir_from_pcs(node, container))

                for logfile in logfiles:
                    log_msg = container_ops.log_random_msg(node,
                                                           container,
                                                           logfile)
                    if pcs_logdir:
                        node_logfile = (pcs_logdir +
                                        f'/{logfile.split("/")[-1]}')
                    else:
                        node_logfile = '/var/log/containers/'\
                                       f'{logfile.split("/")[-2]}/'\
                                       f'{logfile.split("/")[-1]}'
                    self.assertTrue(container_ops.find_msg_in_file(
                        node, node_logfile, log_msg))

    def test_neutron_logs_rotate(self):
        groups = ['controller', 'compute', 'networker']
        neutron_nodes = container_ops.get_nodes_for_groups(groups)
        msg = ''
        for node in neutron_nodes:
            node_logfiles = []
            # set is used to remove duplicated containers
            containers = set(container_ops.get_node_neutron_containers(node) +
                             container_ops.get_node_ovn_containers(node))
            pcs_logdir_dict = {}
            for container in containers:
                cont_logfiles = (container_ops.
                                 get_container_logfiles(node, container))
                if not cont_logfiles:
                    LOG.warning(f'No logfiles have been found in {container} '
                                f'container of {node.name} node')
                    continue
                node_logfiles += cont_logfiles
                # logdir is obtained differently for pcs resources
                pcs_logdir = (container_ops.
                              get_node_logdir_from_pcs(node, container))

                for logfile in cont_logfiles:
                    pcs_logdir_dict[logfile] = pcs_logdir
                    if not msg:
                        msg = container_ops.log_random_msg(node,
                                                           container,
                                                           logfile)
                    else:
                        container_ops.log_msg(node, container, logfile, msg)
            container_ops.rotate_logs(node)
            for logfile in set(node_logfiles):
                if pcs_logdir_dict.get(logfile):
                    node_logfile = (pcs_logdir_dict[logfile] +
                                    f'/{logfile.split("/")[-1]}')
                else:
                    node_logfile = '/var/log/containers/'\
                                   f'{logfile.split("/")[-2]}/'\
                                   f'{logfile.split("/")[-1]}'
                self.assertTrue(container_ops.find_msg_in_file(node,
                                                               node_logfile,
                                                               msg,
                                                               rotated=True))



#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


import subprocess
import os

from oslo_log import log
import podman
import podman1


import tobiko
from tobiko.podman import _exception
from tobiko.podman import _shell
from tobiko.shell import ssh
from tobiko.shell import sh

LOG = log.getLogger(__name__)


def get_podman_client(ssh_client=None):
    return PodmanClientFixture(ssh_client=ssh_client)


def list_podman_containers(client=None, **kwargs):
    try:
        containers = podman_client(client).containers.list(**kwargs)
    except _exception.PodmanSocketNotFoundError:
        return tobiko.Selection()
    else:
        return tobiko.select(containers)


PODMAN_CLIENT_CLASSES = \
    podman1.Client, podman.PodmanClient  # pylint: disable=E1101


def podman_client(obj=None):
    if obj is None:
        obj = get_podman_client()

    if tobiko.is_fixture(obj):
        obj = tobiko.setup_fixture(obj).client

    if isinstance(obj, PODMAN_CLIENT_CLASSES):
        return obj

    raise TypeError('Cannot obtain a Podman client from {!r}'.format(obj))


def podman_version_3():
    try:
        stdout = sh.execute('rpm -q podman').stdout
    except sh.ShellCommandFailed:
        return False

    podman_ver = stdout.split('-')[1].split('.')[0]
    if int(podman_ver) >= 3:
        return True
    else:
        return False


class PodmanClientFixture(tobiko.SharedFixture):

    client = None
    ssh_client = None

    def __init__(self, ssh_client=None):
        super(PodmanClientFixture, self).__init__()
        if ssh_client:
            self.ssh_client = ssh_client

    def setup_fixture(self):
        self.setup_ssh_client()
        self.setup_client()

    def setup_ssh_client(self):
        ssh_client = self.ssh_client
        if ssh_client is None:
            self.ssh_client = ssh_client = ssh.ssh_proxy_client() or False
            if ssh_client:
                tobiko.setup_fixture(ssh_client)
        return ssh_client

    def setup_client(self):
        # podman ver3 (osp>=16.2) has different service / socket paths
        if podman_version_3():
            podman_service = 'podman.socket'
            podman_socket_file = '/run/podman/podman.sock'
        else:
            podman_service = 'io.podman.socket'
            podman_socket_file = '/run/podman/io.podman'

        podman_client_setup_cmds = \
            f"""sudo test -f /var/podman_client_access_setup ||  \
            (sudo groupadd -f podman &&  \
            sudo usermod -a -G podman heat-admin && \
            sudo chmod -R o=wxr /etc/tmpfiles.d && \
            sudo echo 'd /run/podman 0770 root heat-admin' >  \
            /etc/tmpfiles.d/podman.conf && \
            sudo cp /lib/systemd/system/{podman_service} \
            /etc/systemd/system/{podman_service} && \
            sudo crudini --set /etc/systemd/system/{podman_service} Socket  \
            SocketMode 0660 && \
            sudo crudini --set /etc/systemd/system/{podman_service} Socket  \
            SocketGroup podman && \
            sudo systemctl daemon-reload && \
            sudo systemd-tmpfiles --create && \
            sudo systemctl enable --now {podman_service} && \
            sudo chmod 777 /run/podman && \
            sudo chown -R root: /run/podman && \
            sudo chmod g+rw {podman_socket_file} && \
            sudo chmod 777 {podman_socket_file} && \
            sudo setenforce 0 && \
            sudo systemctl start {podman_service} && \
            sudo touch /var/podman_client_access_setup)"""

        sh.execute(podman_client_setup_cmds, ssh_client=self.ssh_client)

        client = self.client
        if client is None:
            self.client = client = self.create_client()
        return client

    def create_client(self):  # noqa: C901
        for _ in tobiko.retry(timeout=60., interval=5.):
            try:
                podman_remote_socket = self.discover_podman_socket()
                username = self.ssh_client.connect_parameters['username']
                host = self.ssh_client.connect_parameters["hostname"]
                socket = podman_remote_socket
                podman_remote_socket_uri = f'unix:/tmp/podman.sock_{host}'

                remote_uri = f'ssh://{username}@{host}{socket}'

                if podman_version_3():
                    # check if a ssh tunnel exists, if not create one
                    psall = str(subprocess.check_output(('ps', '-ef')))
                    if f'ssh -L /tmp/podman.sock_{host}' not in psall:
                        if os.path.exists(f"/tmp/podman.sock_{host}"):
                            subprocess.call(
                                ['rm', '-f', f'/tmp/podman.sock_{host}'])
                        # start a background  ssh tunnel with the remote host
                        subprocess.call(['ssh', '-L',
                                         f'/tmp/podman.sock_{host}:'
                                         f'/run/podman/podman.sock',
                                         host, '-N', '-f'])
                        for _ in tobiko.retry(timeout=60., interval=1.):
                            if os.path.exists(f'/tmp/podman.sock_{host}'):
                                break
                    client = podman.PodmanClient(
                        base_url=podman_remote_socket_uri)
                    if client.ping():
                        LOG.info('container_client is online')

                else:
                    client = podman1.Client(  # pylint: disable=E1101
                        uri=podman_remote_socket_uri,
                        remote_uri=remote_uri,
                        identity_file='~/.ssh/id_rsa')
                    if client.system.ping():
                        LOG.info('container_client is online')
                return client
            except (ConnectionRefusedError, ConnectionResetError):
                # retry
                self.create_client()

    def connect(self):
        return tobiko.setup_fixture(self).client

    def discover_podman_socket(self):
        return _shell.discover_podman_socket(ssh_client=self.ssh_client)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import inspect
import typing

import decorator


def protocol(cls: type) -> type:
    name = cls.__name__
    bases = inspect.getmro(cls)[1:]
    namespace = dict(cls.__dict__,
                     _is_protocol=True,
                     __module__=cls.__module__)
    return type(name, bases, namespace)


def is_protocol_class(cls):
    return inspect.isclass(cls) and cls.__dict__.get('_is_protocol', False)


def is_public_function(obj):
    return (inspect.isfunction(obj) and
            getattr(obj, '__name__', '_')[0] != '_')


T = typing.TypeVar('T')


class CallHandlerMeta(type):

    def __new__(mcls, name, bases, namespace, **kwargs):
        protocol_class = namespace.get('protocol_class')
        if protocol_class is not None:
            proxy_class = call_proxy_class(protocol_class)
            bases += proxy_class,
        return super().__new__(mcls, name, bases, namespace, **kwargs)


class CallHandler(metaclass=CallHandlerMeta):

    protocol_class: type

    def __init__(self,
                 handle_call: typing.Optional[typing.Callable] = None):
        if handle_call is not None:
            assert callable(handle_call)
            setattr(self, '_handle_call', handle_call)

    def _handle_call(self, method: typing.Callable, *args, **kwargs):
        pass

    def use_as(self, cls: typing.Type[T]) -> T:
        assert isinstance(self, cls)
        return typing.cast(T, self)


def call_proxy_class(
        cls: type,
        *bases: type,
        class_name: typing.Optional[str] = None,
        namespace: typing.Optional[dict] = None) \
        -> type:
    if not inspect.isclass(cls):
        raise TypeError(f"Object {cls} is not a class")
    if class_name is None:
        class_name = cls.__name__ + 'Proxy'
    protocol_classes = list_protocols(cls)
    if not protocol_classes:
        raise TypeError(f"Class {cls} doesn't implement any protocol")
    if namespace is None:
        namespace = {}
    for protocol_class in reversed(protocol_classes):
        for name, member in protocol_class.__dict__.items():
            if is_public_function(member):
                method = call_proxy_method(member)
                namespace[name] = method
    # Skip empty protocols
    if not namespace:
        raise TypeError(f"Class {cls} has any protocol specification")
    namespace['__module__'] = cls.__module__
    proxy_class = type(class_name, bases + protocol_classes, namespace)
    assert not is_protocol_class(proxy_class)
    assert not is_protocol_class(proxy_class)
    return proxy_class


def call_proxy(cls: type, handle_call: typing.Callable) -> CallHandler:
    proxy_class = call_proxy_class(cls, CallHandler)
    return proxy_class(handle_call)


@functools.lru_cache()
def list_protocols(cls: type) -> typing.Tuple[type, ...]:
    subclasses = inspect.getmro(cls)
    protocols = tuple(cls
                      for cls in subclasses
                      if is_protocol_class(cls))
    return tuple(protocols)


def call_proxy_method(func: typing.Callable) -> typing.Callable:
    method = decorator.decorate(func, _call_proxy_method)
    assert method is not func
    return method


def _call_proxy_method(func, self: CallHandler, *args, **kwargs):
    # pylint: disable=protected-access
    return self._handle_call(func, *args, **kwargs)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import sys
import typing

from oslo_log import log
import testtools

from tobiko.common import _exception
from tobiko.common import _time


LOG = log.getLogger(__name__)

os.environ.setdefault('PYTHON', sys.executable)


class TestCaseEntry(typing.NamedTuple):
    test_case: testtools.TestCase
    start_time: float


class TestCasesManager(object):

    start_time: _time.Seconds = None

    def __init__(self):
        self._test_cases: typing.List[TestCaseEntry] = []

    def get_test_case(self) -> testtools.TestCase:
        try:
            return self._test_cases[-1].test_case
        except IndexError:
            return DUMMY_TEST_CASE

    def pop_test_case(self) -> testtools.TestCase:
        entry = self._test_cases.pop()
        elapsed_time = _time.time() - entry.start_time
        LOG.debug(f"Exit test case '{entry.test_case.id()}' after "
                  f"{elapsed_time} seconds")
        return entry.test_case

    def push_test_case(self, test_case: testtools.TestCase):
        _exception.check_valid_type(test_case, testtools.TestCase)
        entry = TestCaseEntry(test_case=test_case,
                              start_time=_time.time())
        self._test_cases.append(entry)
        LOG.debug(f"Enter test case '{test_case.id()}'")


TEST_CASES = TestCasesManager()


def push_test_case(test_case: testtools.TestCase,
                   manager: TestCasesManager = TEST_CASES):
    return manager.push_test_case(test_case=test_case)


def pop_test_case(manager: TestCasesManager = TEST_CASES) -> \
        testtools.TestCase:
    return manager.pop_test_case()


def get_test_case(manager: TestCasesManager = TEST_CASES) -> \
        testtools.TestCase:
    return manager.get_test_case()


class DummyTestCase(testtools.TestCase):

    def runTest(self):
        pass


DUMMY_TEST_CASE = DummyTestCase()


def run_test(test_case: testtools.TestCase,
             test_result: testtools.TestResult = None) -> testtools.TestResult:
    test_result = test_result or testtools.TestResult()
    test_case.run(test_result)
    return test_result


def assert_in(needle, haystack, message: typing.Optional[str] = None,
              manager: TestCasesManager = TEST_CASES):
    get_test_case(manager=manager).assertIn(needle, haystack, message)


def get_skipped_test_cases(test_result: testtools.TestResult,
                           skip_reason: typing.Optional[str] = None):
    if skip_reason is not None:
        assert_in(skip_reason, test_result.skip_reasons)
        return test_result.skip_reasons[skip_reason]
    else:
        skipped_test_cases = list()
        for cases in test_result.skip_reasons.values():
            skipped_test_cases.extend(cases)
        return skipped_test_cases


def assert_test_case_was_skipped(test_case: testtools.TestCase,
                                 test_result: testtools.TestResult,
                                 skip_reason: str = None,
                                 manager: TestCasesManager = TEST_CASES):
    skipped_tests = get_skipped_test_cases(test_result=test_result,
                                           skip_reason=skip_reason)
    assert_in(test_case, skipped_tests, manager=manager)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from octaviaclient.api.v2 import octavia

import tobiko
from tobiko.openstack import _client
from tobiko.openstack import keystone


OCTAVIA_CLIENT_CLASSSES = octavia.OctaviaAPI,


def get_octavia_endpoint(keystone_client=None):
    return keystone.find_service_endpoint(name='octavia',
                                          client=keystone_client)


class OctaviaClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session):
        keystone_client = keystone.get_keystone_client(session=session)
        endpoint = get_octavia_endpoint(keystone_client=keystone_client)
        return octavia.OctaviaAPI(session=session, endpoint=endpoint.url)


class OctaviaClientManager(_client.OpenstackClientManager):

    def create_client(self, session):
        return OctaviaClientFixture(session=session)


CLIENTS = OctaviaClientManager()


def octavia_client(obj):
    if not obj:
        return get_octavia_client()

    if isinstance(obj, OCTAVIA_CLIENT_CLASSSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, OctaviaClientFixture):
        return fixture.client

    message = "Object {!r} is not an OctaviaClientFixture".format(obj)
    raise TypeError(message)


def get_octavia_client(session=None, shared=True, init_client=None,
                       manager=None):
    manager = manager or CLIENTS
    client = manager.get_client(session=session, shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client


def get_loadbalancer(loadbalancer_id, client=None):
    return octavia_client(client).load_balancer_show(lb_id=loadbalancer_id)


def get_member(pool_id, member_id, client=None):
    return octavia_client(client).member_show(pool_id=pool_id,
                                              member_id=member_id)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

import mock
import testtools

import tobiko
from tobiko.tests import unit


class RetryTest(unit.TobikoUnitTest):

    def test_retry_when_succeed(self,):
        mock_time = self.patch_time()
        attempts = []

        for attempt in tobiko.retry():
            attempts.append(attempt)
            break  # this marks a success

        expected = [tobiko.retry_attempt(number=1,
                                         start_time=mock_time.start_time,
                                         elapsed_time=0.)]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_not_called()

    def test_retry_untill_succeed(self):
        mock_time = self.patch_time()
        attempts = []

        for attempt in tobiko.retry():
            attempts.append(attempt)
            if attempt.number >= 3:
                break

        time_increment = mock_time.time_increment
        expected = [tobiko.retry_attempt(number=i+1,
                                         start_time=mock_time.start_time,
                                         elapsed_time=i * time_increment)
                    for i in range(3)]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_not_called()

    def test_retry_with_count(self):
        mock_time = self.patch_time()
        attempts = []

        try:
            for attempt in tobiko.retry(count=2):
                attempts.append(attempt)
        except tobiko.RetryCountLimitError as ex:
            self.assertEqual("Retry count limit exceeded "
                             f"({attempt.details})", str(ex))
        else:
            self.fail("RetryCountLimitError not raised")

        time_increment = mock_time.time_increment
        expected = [tobiko.retry_attempt(number=i+1,
                                         count=2,
                                         start_time=mock_time.start_time,
                                         elapsed_time=i * time_increment)
                    for i in range(2)]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_not_called()

    def test_retry_with_timeout(self):
        mock_time = self.patch_time()
        attempts = []

        try:
            for attempt in tobiko.retry(timeout=2.5):
                attempts.append(attempt)
        except tobiko.RetryTimeLimitError as ex:
            self.assertEqual("Retry time limit exceeded "
                             f"({attempt.details})", str(ex))
        else:
            self.fail("RetryTimeLimitError not raised")

        time_increment = mock_time.time_increment
        expected = [tobiko.retry_attempt(number=i+1,
                                         timeout=2.5,
                                         start_time=mock_time.start_time,
                                         elapsed_time=i * time_increment)
                    for i in range(4)]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_not_called()

    def test_retry_with_interval(self):
        mock_time = self.patch_time()
        attempts = []

        for attempt in tobiko.retry(interval=5.):
            attempts.append(attempt)
            if attempt.number >= 3:
                break

        expected = [tobiko.retry_attempt(number=i+1,
                                         interval=5.,
                                         start_time=mock_time.start_time,
                                         elapsed_time=elapsed_time)
                    for i, elapsed_time in enumerate([0., 6., 11.])]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_has_calls([mock.call(4.),
                                          mock.call(3.)])

    def test_retry_with_timeout_and_small_interval(self):
        mock_time = self.patch_time()
        attempts = []

        try:
            for attempt in tobiko.retry(timeout=2.5, interval=0.5):
                attempts.append(attempt)
        except tobiko.RetryTimeLimitError as ex:
            self.assertEqual("Retry time limit exceeded "
                             f"({attempt.details})", str(ex))
        else:
            self.fail("RetryTimeLimitError not raised")

        expected = [tobiko.retry_attempt(number=i+1,
                                         timeout=2.5,
                                         interval=0.5,
                                         start_time=mock_time.start_time,
                                         elapsed_time=elapsed_time)
                    for i, elapsed_time in enumerate([0., 1., 2., 3.])]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_not_called()

    def test_retry_with_timeout_and_big_interval(self):
        mock_time = self.patch_time()
        attempts = []

        try:
            for attempt in tobiko.retry(timeout=9., interval=3.):
                attempts.append(attempt)
        except tobiko.RetryTimeLimitError as ex:
            self.assertEqual("Retry time limit exceeded "
                             f"({attempt.details})", str(ex))
        else:
            self.fail("RetryTimeLimitError not raised")

        expected = [tobiko.retry_attempt(number=i+1,
                                         timeout=9.,
                                         interval=3.,
                                         start_time=mock_time.start_time,
                                         elapsed_time=elapsed_time)
                    for i, elapsed_time in enumerate([0., 4., 7., 10.])]
        self.assertEqual(expected, attempts)
        mock_time.sleep.assert_has_calls([mock.call(2.),
                                          mock.call(1.),
                                          mock.call(1.)])

    def test_retry_on_exception_when_succeed(self):
        count_calls = itertools.count()

        @tobiko.retry_on_exception(ValueError)
        def func(a, b):
            next(count_calls)
            return a + b

        result = func(3, 4)
        self.assertEqual(1, next(count_calls))
        self.assertEqual(7, result)

    def test_retry_on_exception_when_fails(self):
        count_calls = itertools.count()

        @tobiko.retry_on_exception(ValueError)
        def func(a, b):
            next(count_calls)
            raise RuntimeError(f"{a + b}")

        ex = self.assertRaises(RuntimeError, func, 3, 4)
        self.assertEqual(1, next(count_calls))
        self.assertEqual('7', str(ex))

    def test_retry_on_exception_untill_succeed(self):
        count_calls = itertools.count()

        @tobiko.retry_on_exception(ValueError, TypeError)
        def func(a, b):
            count = next(count_calls)
            if count == 0:
                raise ValueError
            if count == 1:
                raise TypeError
            return a + b

        result = func(3, 4)
        self.assertEqual(3, next(count_calls))
        self.assertEqual(7, result)

    def test_retry_on_exception_with_count(self):
        count_calls = itertools.count()

        @tobiko.retry_on_exception(RuntimeError, count=3)
        def func(_a, _b):
            count = next(count_calls)
            raise RuntimeError(f"{count}")

        ex = self.assertRaises(RuntimeError, func, 3, 4)
        self.assertEqual(3, next(count_calls))
        self.assertEqual("2", str(ex))

    def test_retry_on_exception_with_timeout(self):
        mock_time = self.patch_time()
        count_calls = itertools.count()

        @tobiko.retry_on_exception(ValueError, timeout=3.)
        def func(_a, _b):
            count = next(count_calls)
            raise ValueError(f"{count}")

        ex = self.assertRaises(ValueError, func, 3, 4)
        self.assertEqual(4, next(count_calls))
        self.assertEqual("3", str(ex))
        self.assertEqual(4., mock_time.current_time)
        mock_time.sleep.assert_not_called()

    def test_retry_on_exception_with_interval(self):
        mock_time = self.patch_time()
        count_calls = itertools.count()

        @tobiko.retry_on_exception(ValueError, interval=5.)
        def func(a, b):
            count = next(count_calls)
            if count > 2:
                return a + b
            raise ValueError(f"{count}")

        result = func(3, 4)
        self.assertEqual(4, next(count_calls))
        self.assertEqual(7, result)
        self.assertEqual(17., mock_time.current_time)
        mock_time.sleep.assert_has_calls([mock.call(4.),
                                          mock.call(3.),
                                          mock.call(3.)])

    def test_retry_test_case_when_succeed(self):

        class MyTest(testtools.TestCase):

            @tobiko.retry_test_case()
            def test_success(self):
                pass

        result = testtools.TestResult()
        test_case = MyTest('test_success')
        test_case.run(result)

        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.failures)
        self.assertEqual([], result.errors)
        self.assertEqual({}, result.skip_reasons)

    def test_retry_test_case_when_fails(self):

        class MyTest(testtools.TestCase):

            @tobiko.retry_test_case()
            def test_failure(self):
                try:
                    self.fail("this is failing")
                except tobiko.FailureException as ex:
                    failures.append(ex)
                    raise

        failures = []
        result = testtools.TestResult()
        test_case = MyTest('test_failure')
        test_case.run(result)

        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.errors)
        self.assertEqual({}, result.skip_reasons)

        self.assertEqual(3, len(failures))
        self.assertEqual(1, len(result.failures))
        failed_test_case, traceback = result.failures[0]
        self.assertIs(test_case, failed_test_case)
        self.assertIn(str(failures[-1]), traceback)

    def test_retry_test_case_when_fails_once(self):

        class MyTest(testtools.TestCase):

            @tobiko.retry_test_case()
            def test_one_failure(self):
                count = next(count_calls)
                self.assertNotEqual(0, count)

        count_calls = itertools.count()
        result = testtools.TestResult()
        test_case = MyTest('test_one_failure')
        test_case.run(result)

        self.assertEqual(2, next(count_calls))
        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.failures)
        self.assertEqual([], result.errors)
        self.assertEqual({}, result.skip_reasons)

    def test_retry_test_case_when_raises_errors(self):

        class MyTest(testtools.TestCase):

            @tobiko.retry_test_case()
            def test_errors(self):
                ex = ValueError('pippo')
                errors.append(ex)
                raise ex

        errors = []
        result = testtools.TestResult()
        test_case = MyTest('test_errors')
        test_case.run(result)

        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.failures)
        self.assertEqual({}, result.skip_reasons)

        self.assertEqual(1, len(errors))
        self.assertEqual(1, len(result.errors))
        failed_test_case, traceback = result.errors[0]
        self.assertIs(test_case, failed_test_case)
        self.assertIn(str(errors[-1]), traceback)

    def test_retry_test_case_when_skip(self):

        class MyTest(testtools.TestCase):

            @tobiko.retry_test_case()
            def test_skip(self):
                next(count_calls)
                self.skip("Not the right day!")

        count_calls = itertools.count()
        result = testtools.TestResult()
        test_case = MyTest('test_skip')
        test_case.run(result)

        self.assertEqual(1, next(count_calls))
        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.failures)
        self.assertEqual([], result.errors)
        self.assertEqual({"Not the right day!": [test_case]},
                         result.skip_reasons)

    def test_attempt_is_last_with_count(self):
        self.assertFalse(tobiko.retry_attempt(number=2,
                                              count=3).is_last)
        self.assertTrue(tobiko.retry_attempt(number=2,
                                             count=2).is_last)

    def test_attempt_is_last_with_timeout(self):
        self.assertFalse(tobiko.retry_attempt(elapsed_time=2.,
                                              timeout=3.).is_last)
        self.assertTrue(tobiko.retry_attempt(elapsed_time=2.,
                                             timeout=2.).is_last)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import enum

from oslo_log import log
import six

import tobiko
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _process


LOG = log.getLogger(__name__)


DATA_TYPES = six.string_types + (six.binary_type, six.text_type)


@enum.unique
class ShellExecuteStatus(enum.Enum):
    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    TIMEDOUT = 'TIMEDOUT'
    UNTERMINATED = 'UNTERMINATED'


def execute_result(command, exit_status=None, timeout=None,
                   status=None, login=None, stdin=None, stdout=None,
                   stderr=None):
    command = str(command)
    if exit_status is not None:
        exit_status = int(exit_status)
    if timeout is not None:
        timeout = float(timeout)
    if status is not None:
        status = ShellExecuteStatus(status)
    stdin = _process.str_from_stream(stdin)
    stdout = _process.str_from_stream(stdout)
    stderr = _process.str_from_stream(stderr)
    return ShellExecuteResult(command=command,
                              exit_status=exit_status,
                              timeout=timeout,
                              status=status,
                              stdin=stdin,
                              stdout=stdout,
                              stderr=stderr,
                              login=login)


class ShellExecuteResult(collections.namedtuple(
        'ShellExecuteResult', ['command', 'exit_status', 'timeout', 'status',
                               'login', 'stdin', 'stdout', 'stderr'])):

    _details = None

    @property
    def details(self):
        details = self._details
        if details is None:
            self._details = details = self.get_details()
        return details

    def get_details(self):
        details = []
        details.append("command: {!r}".format(self.command))

        exit_status = self.exit_status
        if exit_status is not None:
            details.append("exit_status: {!r}".format(exit_status))

        timeout = self.timeout
        if timeout is not None:
            details.append("timeout: {!r}".format(timeout))

        status = self.status
        if status is not None:
            details.append("status: {!s}".format(status))

        login = self.login
        if login is not None:
            details.append("login: {!r}".format(login))

        stdin = self.stdin
        if stdin:
            details.append("stdin:\n{!s}".format(_indent(stdin)))

        stdout = self.stdout
        if stdout:
            details.append("stdout:\n{!s}".format(_indent(stdout)))

        stderr = self.stderr
        if stderr:
            details.append("stderr:\n{!s}".format(_indent(stderr)))
        return '\n'.join(details)

    def format(self):
        return self.details


def _indent(text, space='    ', newline='\n'):
    text = str(text)
    return space + (newline + space).join(text.split(newline))


def execute(command, environment=None, timeout=None, shell=None,
            stdin=None, stdout=None, stderr=None, ssh_client=None,
            expect_exit_status=0, **kwargs):
    """Execute command inside a remote or local shell

    :param command: command argument list

    :param timeout: command execution timeout in seconds

    :param check: when False it doesn't raises ShellCommandError when
    exit status is not zero. True by default

    :param ssh_client: SSH client instance used for remote shell execution

    :raises ShellTimeoutExpired: when timeout expires before command execution
    terminates. In such case it kills the process, then it eventually would
    try to read STDOUT and STDERR buffers (not fully implemented) before
    raising the exception.

    :raises ShellCommandError: when command execution terminates with non-zero
    exit status.
    """
    process = _process.process(command=command,
                               environment=environment,
                               timeout=timeout,
                               shell=shell,
                               stdin=stdin,
                               stdout=stdout,
                               stderr=stderr,
                               ssh_client=ssh_client,
                               **kwargs)
    login = getattr(ssh_client, 'login', None)
    return execute_process(process=process,
                           stdin=stdin,
                           login=login,
                           expect_exit_status=expect_exit_status)


def execute_process(process, stdin, expect_exit_status, login=None):
    error = None
    status = None
    try:
        with process:
            if stdin and isinstance(stdin, DATA_TYPES):
                process.send_all(data=stdin)
    except _exception.ShellTimeoutExpired:
        status = ShellExecuteStatus.TIMEDOUT
        if expect_exit_status is not None:
            error = tobiko.exc_info()
    else:
        if expect_exit_status is not None:
            try:
                process.check_exit_status(expect_exit_status)
            except _exception.ShellCommandFailed:
                status = ShellExecuteStatus.FAILED
                error = tobiko.exc_info()
            except _exception.ShellProcessNotTerminated:
                status = ShellExecuteStatus.UNTERMINATED
            else:
                status = ShellExecuteStatus.SUCCEEDED

    result = execute_result(command=process.command,
                            exit_status=process.exit_status,
                            timeout=process.timeout,
                            status=status,
                            login=login,
                            stdin=process.stdin,
                            stdout=process.stdout,
                            stderr=process.stderr)
    if error:
        LOG.info("Command error:\n%s\n", result.details)
        error.result = result
        error.reraise()

    LOG.debug("Command executed:\n%s\n", result.details)
    return result



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import typing

from oslo_log import log

import tobiko
from tobiko.openstack.nova import _client


LOG = log.getLogger(__name__)


class NovaServiceException(tobiko.TobikoException):
    pass


class NovaServicesNotfound(NovaServiceException):
    message = "Nova services not found ({attributes})"


class NovaServicesFailed(NovaServiceException):
    message = "Nova services are failed:\n{details}"


def services_details(services: typing.List):
    # pylint: disable=protected-access
    return json.dumps([service._info for service in services],
                      indent=4, sort_keys=True)


def wait_for_services_up(retry: typing.Optional[tobiko.Retry] = None,
                         **list_services_params):
    for attempt in tobiko.retry(other_retry=retry,
                                default_timeout=300.,
                                default_interval=5.):
        services = _client.list_services(**list_services_params)
        LOG.debug(f"Found {len(services)} Nova services")
        try:
            if not services:
                raise NovaServicesNotfound(
                    attributes=json.dumps(list_services_params))

            heathy_services = services.with_attributes(state='up')
            LOG.debug(f"Found {len(heathy_services)} healthy Nova services")

            failed_services = [service
                               for service in services
                               if service not in heathy_services]
            LOG.debug(f"Found {len(failed_services)} failed Nova services")
            if failed_services:
                details = services_details(failed_services)
                LOG.info(f"Failed Nova services:\n{details}")
                raise NovaServicesFailed(details=details)
            LOG.info('All nova services are up!')
            break  # all Nova services are healthy

        except NovaServiceException:
            # Re-raises this exception in case this is the last retry
            # attempt
            attempt.check_limits()
            continue



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class NetworkTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_setup_fixture(stacks.NetworkStackFixture)

    @neutron.skip_if_missing_networking_extensions('port-security')
    def test_port_security_enabled(self):
        self.assertEqual(self.stack.port_security_enabled,
                         self.stack.network_details['port_security_enabled'])
        self.assertEqual(self.stack.port_security_enabled,
                         self.stack.outputs.port_security_enabled)

    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_net_mtu(self):
        self.assertEqual(self.stack.network_details['mtu'],
                         self.stack.outputs.mtu)

    def test_ipv4_subnet_cidr(self):
        if not self.stack.has_ipv4:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no ipv4 "
                             "subnet")

        subnet = neutron.find_subnet(cidr=str(self.stack.ipv4_subnet_cidr))
        self.assertEqual(neutron.get_subnet(self.stack.ipv4_subnet_id), subnet)

    def test_ipv6_subnet_cidr(self):
        if not self.stack.has_ipv6:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no ipv6 "
                             "subnet")
        subnet = neutron.find_subnet(cidr=str(self.stack.ipv6_subnet_cidr))
        self.assertEqual(neutron.get_subnet(self.stack.ipv6_subnet_id), subnet)

    def test_gateway_network(self):
        if not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no gateway")
        self.assertEqual(
            self.stack.gateway_network_id,
            self.stack.gateway_details['external_gateway_info']['network_id'])

    def test_ipv4_subnet_gateway_ip(self):
        if not self.stack.has_ipv4 or not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no IPv4 "
                             "gateway")
        self.assertIn(
            self.stack.ipv4_subnet_gateway_ip,
            self.stack.ipv4_gateway_addresses)

    def test_ipv6_subnet_gateway_ip(self):
        if not self.stack.has_ipv6 or not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no IPv6 "
                             "gateway")
        self.assertIn(
            self.stack.ipv6_subnet_gateway_ip,
            self.stack.ipv6_gateway_addresses)


@keystone.skip_unless_has_keystone_credentials()
@neutron.skip_if_missing_networking_extensions('net-mtu-write')
class NetworkWithNetMtuWriteTest(NetworkTest):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_setup_fixture(
        stacks.NetworkWithNetMtuWriteStackFixture)

    def test_net_mtu_write(self):
        self.assertEqual(self.stack.mtu, self.stack.outputs.mtu)


@keystone.skip_unless_has_keystone_credentials()
@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(neutron.L3_AGENT, 2)
class L3HaNetworkTest(NetworkTest):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_setup_fixture(stacks.L3haNetworkStackFixture)


@keystone.skip_unless_has_keystone_credentials()
@stacks.skip_unless_has_external_network
class ExternalNetworkStackTest(testtools.TestCase):

    def test_get_external_network(self):
        network = stacks.get_external_network()
        self.assertTrue(network['id'])
        self.assertIs(True, network['router:external'])
        self.assertEqual('ACTIVE', network['status'])
        subnets = neutron.list_subnets(network_id=network['id'])
        self.assertNotEqual([], subnets)

    def test_has_external_network(self):
        self.assertIs(True, stacks.has_external_network())


@keystone.skip_unless_has_keystone_credentials()
@stacks.skip_unless_has_floating_network
class FloatingNetworkStackTest(testtools.TestCase):

    def test_get_floating_network(self):
        network = stacks.get_floating_network()
        self.assertTrue(network['id'])
        self.assertIs(True, network['router:external'])
        self.assertEqual('ACTIVE', network['status'])

    def test_has_floating_network(self):
        self.assertIs(True, stacks.has_floating_network())



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import contextlib
import typing

from oslo_log import log
import yaml

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh

LOG = log.getLogger(__name__)


CLOUD_INIT_TRANSIENT_STATES = {
    'done': tuple(['running'])
}

CLOUD_INIT_OUTPUT_FILE = '/var/log/cloud-init-output.log'
CLOUD_INIT_LOG_FILE = '/var/log/cloud-init.log'


def user_data(*args, **kwargs):
    config = cloud_config(*args, **kwargs)
    if config:
        return config.user_data
    else:
        return ''


def cloud_config(*args, **kwargs):
    return combine_cloud_configs(args + (kwargs,))


def combine_cloud_configs(objs):
    packages = []
    runcmd = []
    extra_params = {}
    for obj in objs:
        if obj:
            if not isinstance(obj, collections.abc.Mapping):
                obj = dict(obj)
            for package in obj.pop('packages', []):
                if package and package not in packages:
                    packages.append(package)
            for cmdline in obj.pop('runcmd', []):
                if cmdline:
                    cmdline = list(sh.shell_command(cmdline))
                    if cmdline:
                        runcmd.append(cmdline)
            if obj:
                extra_params.update(obj)

    return CloudConfig.create(packages=packages or None,
                              runcmd=runcmd or None,
                              **extra_params)


class CloudConfig(dict):

    @classmethod
    def create(cls, **kwargs):
        return cls((k, v)
                   for k, v in kwargs.items()
                   if v is not None)

    @property
    def user_data(self):
        return '#cloud-config\n' + yaml.dump(dict(self))

    def __add__(self, other):
        return combine_cloud_configs([self, other])


class InvalidCloudInitStatusError(tobiko.TobikoException):
    message = ("cloud-init status of host '{hostname}' is "
               "'{actual_status}' while it is expecting to "
               "be in {expected_states!r}:\n\n"
               f"--- {CLOUD_INIT_LOG_FILE} ---\n"
               "{log_file}\n\n"
               f"--- {CLOUD_INIT_OUTPUT_FILE} ---\n"
               "{output_file}\n\n")


class WaitForCloudInitTimeoutError(InvalidCloudInitStatusError):
    message = ("after {timeout} seconds cloud-init status of host "
               "'{hostname}' is still '{actual_status}' while it is "
               "expecting to be in {expected_states!r}:\n\n"
               f"--- {CLOUD_INIT_LOG_FILE} ---\n"
               "{log_file}\n\n"
               f"--- {CLOUD_INIT_OUTPUT_FILE} ---\n"
               "{output_file}\n\n")


def get_cloud_init_status(
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None) \
        -> str:
    try:
        output = sh.execute('cloud-init status',
                            ssh_client=ssh_client,
                            timeout=timeout,
                            sudo=True).stdout
    except sh.ShellCommandFailed as ex:
        output = ex.stdout
        if output:
            LOG.debug(f"Cloud init status error reported:\n{ex}")
        else:
            raise

    status = yaml.load(output)
    tobiko.check_valid_type(status, dict)
    tobiko.check_valid_type(status['status'], str)
    return status['status']


def wait_for_cloud_init_done(
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None,
        sleep_interval: tobiko.Seconds = None) \
        -> str:
    return wait_for_cloud_init_status('done',
                                      ssh_client=ssh_client,
                                      timeout=timeout,
                                      sleep_interval=sleep_interval)


def wait_for_cloud_init_status(
        *expected_states: str,
        transient_states: typing.Optional[typing.Container[str]] = None,
        ssh_client: typing.Optional[ssh.SSHClientFixture] = None,
        timeout: tobiko.Seconds = None,
        sleep_interval: tobiko.Seconds = None) \
        -> str:
    hostname = getattr(ssh_client, 'hostname', None)
    if transient_states is None:
        transient_states = list()
        for status in expected_states:
            transient_states += CLOUD_INIT_TRANSIENT_STATES.get(status, [])

    with open_output_file(filename=CLOUD_INIT_LOG_FILE,
                          timeout=timeout,
                          ssh_client=ssh_client) as log_file, \
            open_output_file(filename=CLOUD_INIT_OUTPUT_FILE,
                             timeout=timeout,
                             ssh_client=ssh_client) as output_file:
        for attempt in tobiko.retry(timeout=timeout,
                                    interval=sleep_interval,
                                    default_timeout=1200.,
                                    default_interval=5.):
            actual_status = get_cloud_init_status(ssh_client=ssh_client,
                                                  timeout=attempt.time_left)
            if actual_status in expected_states:
                return actual_status

            log_file.readall()
            output_file.readall()
            if actual_status not in transient_states:
                raise InvalidCloudInitStatusError(
                    hostname=hostname,
                    actual_status=actual_status,
                    expected_states=expected_states,
                    log_file=str(log_file),
                    output_file=str(output_file))

            try:
                attempt.check_limits()
            except tobiko.RetryTimeLimitError as ex:
                raise WaitForCloudInitTimeoutError(
                    timeout=attempt.timeout,
                    hostname=hostname,
                    actual_status=actual_status,
                    expected_states=expected_states,
                    log_file=str(log_file),
                    output_file=str(output_file)) from ex

            # show only the last log line
            try:
                last_log_line = str(log_file).splitlines()[-1]
            except IndexError:
                last_log_line = ""
            LOG.debug(f"Waiting cloud-init status on host '{hostname}' to "
                      f"switch from '{actual_status}' to any of expected "
                      f"states ({', '.join(expected_states)}):\n\n"
                      f"--- {CLOUD_INIT_LOG_FILE} ---\n"
                      f"{last_log_line}\n\n")

    raise RuntimeError("Retry loop ended himself")


@contextlib.contextmanager
def open_output_file(filename: str = CLOUD_INIT_OUTPUT_FILE,
                     tail=False,
                     follow=False,
                     **process_params) \
        -> typing.Generator[sh.ShellStdout, None, None]:
    command = ['tail']
    if not tail:
        # Start from the begin of the file
        command += ['-c', '+0']
    if follow:
        command += ['-F']

    command += [filename]
    process = sh.process(command, **process_params)
    with process:
        yield process.stdout



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.tests import unit


class TestCaseTest(unit.TobikoUnitTest):

    def setUp(self):
        super(TestCaseTest, self).setUp()
        self.addCleanup(self._pop_inner_test_cases)

    def _pop_inner_test_cases(self):
        case = tobiko.get_test_case()
        while case is not self:
            tobiko.pop_test_case()
            case = tobiko.get_test_case()

    def test_get_test_case(self):
        result = tobiko.get_test_case()
        self.assertIs(self, result)

    def test_get_test_case_out_of_context(self):
        manager = tobiko.TestCasesManager()
        result = tobiko.get_test_case(manager=manager)
        self.assertIsInstance(result, testtools.TestCase)
        self.assertEqual('tobiko.common._testcase.DummyTestCase.runTest',
                         result.id())

    def test_push_test_case(self):

        class InnerTest(testtools.TestCase):

            def runTest(self):
                pass

        inner_case = InnerTest()

        tobiko.push_test_case(inner_case)
        self.assertIs(inner_case, tobiko.get_test_case())

    def test_pop_test_case(self):

        class InnerTest(testtools.TestCase):

            def runTest(self):
                pass

        inner_case = InnerTest()
        tobiko.push_test_case(inner_case)

        result = tobiko.pop_test_case()

        self.assertIs(inner_case, result)
        self.assertIs(self, tobiko.get_test_case())



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.shell import ip
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.openstack import topology


class OpenstackNodesTest(testtools.TestCase):

    topology = tobiko.required_setup_fixture(
        topology.get_default_openstack_topology_class())

    def test_public_ips(self):
        ips = dict()
        for node in self.topology.nodes:
            ping.ping(node.public_ip).assert_replied()
            other = ips.setdefault(node.public_ip, node)
            if node is not other:
                tobiko.fail("Nodes {!r} and {!r} have the same IP: {!s}",
                            node.name, other.name, node.public_ip)

    def test_hostnames(self):
        hostnames = dict()
        for node in self.topology.nodes:
            hostname = sh.get_hostname(ssh_client=node.ssh_client)
            self.assertTrue(hostname.startswith(node.name))
            other = hostnames.setdefault(hostname, node)
            if node is not other:
                tobiko.fail("Nodes {!r} and {!r} have the same hostname: {!r}",
                            node.name, other.name, hostname)

    def test_network_namespaces(self):
        for node in self.topology.nodes:
            namespaces_ips = {}
            namespaces = ip.list_network_namespaces(ssh_client=node.ssh_client)
            for namespace in namespaces:
                ips = ip.list_ip_addresses(ssh_client=node.ssh_client,
                                           network_namespace=namespace)
                other_ips = namespaces_ips.setdefault(namespace, ips)
                if ips is not other_ips:
                    tobiko.fail("Duplicate network namespace {!r} in node "
                                "{!r}: {!r}, {!r}", namespace, node.name,
                                other_ips, ips)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.openstack import neutron
from tobiko.openstack import stacks
from tobiko.openstack import topology


LOG = log.getLogger(__name__)
CONF = config.CONF


class RouterTest(testtools.TestCase):
    """Test Neutron routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.CirrosServerStackFixture)

    def setUp(self):
        super(RouterTest, self).setUp()
        if not self.stack.network_stack.has_gateway:
            tobiko.skip_test(
                f"Stack {self.stack.network_stack.stack_name} has no gateway")

    @property
    def router(self):
        return self.stack.network_stack.gateway_details

    @property
    def ipv4_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv4:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv4 subnet")
        return self.stack.network_stack.ipv4_subnet_gateway_ip

    @property
    def ipv6_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv6:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv6 subnet")
        return self.stack.network_stack.ipv6_subnet_gateway_ip

    @property
    def external_gateway_ips(self):
        return self.stack.network_stack.external_gateway_ips

    @property
    def router_ips(self):
        return tobiko.select([self.ipv4_subnet_gateway_ip,
                              self.ipv6_subnet_gateway_ip] +
                             self.external_gateway_ips)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv4_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv4_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv6_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv6_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    def test_ipv4_subnet_gateway_ip(self):
        self.assertEqual(4, self.ipv4_subnet_gateway_ip.version)
        self.assertIn(self.ipv4_subnet_gateway_ip,
                      self.stack.network_stack.ipv4_gateway_addresses)

    def test_ipv6_subnet_gateway_ip(self):
        self.assertEqual(6, self.ipv6_subnet_gateway_ip.version)
        self.assertIn(self.ipv6_subnet_gateway_ip,
                      self.stack.network_stack.ipv6_gateway_addresses)

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        router_agent = neutron.find_l3_agent_hosting_router(self.router['id'],
                                                            unique=True)
        self._check_routers_namespace_on_host(router_agent['host'])

    def _check_routers_namespace_on_host(self, hostname, state="master"):
        router_namespace = "qrouter-%s" % self.router['id']
        agent_host = topology.get_openstack_node(hostname=hostname)
        namespaces = ip.list_network_namespaces(
            ssh_client=agent_host.ssh_client)
        self.assertIn(router_namespace, namespaces)
        namespace_ips = ip.list_ip_addresses(
            scope='global', network_namespace=router_namespace,
            ssh_client=agent_host.ssh_client)
        missing_ips = set(self.router_ips) - set(namespace_ips)
        if state == "master":
            self.assertFalse(missing_ips)
        else:
            self.assertTrue(missing_ips)


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HARouterTest(RouterTest):
    """Test Neutron HA routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_setup_fixture(stacks.L3haServerStackFixture)

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        master_agent, backup_agents = (
            neutron.wait_for_master_and_backup_agents(self.router['id']))
        self.assertGreaterEqual(len(backup_agents), 1)
        self._check_routers_namespace_on_host(master_agent['host'])
        for backup_agent in backup_agents:
            self._check_routers_namespace_on_host(
                backup_agent['host'], state="backup")


class NetworkWithNoServersStack(stacks.NetworkStackFixture):
    pass


@neutron.skip_unless_is_ovs()
class DVRTest(testtools.TestCase):

    router_stack = tobiko.required_setup_fixture(NetworkWithNoServersStack)
    server_stack = tobiko.required_setup_fixture(
            stacks.CirrosServerStackFixture)

    def setUp(self):
        super(DVRTest, self).setUp()
        if not self.router_stack.gateway_details.get('distributed'):
            tobiko.skip_test('No DVR enabled')

    def test_router_not_created_on_compute_if_no_instance_connected(self):
        '''Test that no router namespace is created for DVR on compute node

        Namespace should be only created if there is VM with router that is set
        as a default gateway. Need to verify that there will be no namespace
        created on the compute node where VM is connected to the external
        network. The same network is used as the default gateway for the router
        '''

        router_namespace = f'qrouter-{self.router_stack.gateway_details["id"]}'
        cirros_hypervisor = topology.get_openstack_node(
                hostname=self.server_stack.hypervisor_host)
        namespaces = ip.list_network_namespaces(
            ssh_client=cirros_hypervisor.ssh_client)
        self.assertNotIn(router_namespace, namespaces)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from unittest import mock

import tobiko
from tobiko.tests import unit
from tobiko import actor


@tobiko.protocol
class Greeter:

    async def greet(self, whom: str, greeted: 'Greeted'):
        raise NotImplementedError


@tobiko.protocol
class Greeted:

    def greeted(self, whom: str, greeter: Greeter):
        raise NotImplementedError


class GreeterRef(actor.ActorRef):
    pass


class GreeterActor(Greeter, actor.Actor):

    setup_called = False
    cleanup_called = False
    base_ref_class = GreeterRef

    async def setup_actor(self):
        self.setup_called = True

    async def cleanup_actor(self):
        self.cleanup_called = True

    @actor.actor_method
    async def greet(self, whom: str, greeted: Greeted):
        assert self.setup_called
        assert not self.cleanup_called
        if not whom:
            raise ValueError("'whom' parameter can't be empty")

        self.log.info(f"Hello {whom}!")
        greeted.greeted(whom=whom, greeter=self.ref.use_as(Greeter))


class ActorTest(unit.TobikoUnitTest):

    def test_greeter_ref_class(self):
        ref_class = GreeterActor.ref_class()
        self.assertTrue(issubclass(ref_class, actor.ActorRef))
        self.assertTrue(issubclass(ref_class, GreeterRef))
        self.assertTrue(issubclass(ref_class, Greeter))

    async def test_async_request(self):
        greeter = actor.create_actor(GreeterActor).use_as(Greeter)
        self.assertIsInstance(greeter, actor.ActorRef)
        self.assertIsInstance(greeter, GreeterRef)
        self.assertIsInstance(greeter, Greeter)
        greeted = mock.MagicMock(spec=Greeted)

        await greeter.greet(whom=self.id(), greeted=greeted)
        greeted.greeted.assert_called_with(whom=self.id(),
                                           greeter=greeter)

    async def test_async_request_failure(self):
        greeter = actor.create_actor(GreeterActor).use_as(Greeter)
        self.assertIsInstance(greeter, actor.ActorRef)
        self.assertIsInstance(greeter, Greeter)
        greeted = mock.MagicMock(spec=Greeted)

        try:
            await greeter.greet(whom="", greeted=greeted)
        except ValueError as ex:
            self.assertEqual("'whom' parameter can't be empty", str(ex))
        else:
            self.fail("Exception not raised")
        greeted.greeted.assert_not_called()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import asyncio
import functools
import inspect
import shutil
import os
import tempfile

from oslo_log import log
import testtools

import tobiko
from tobiko.tests.unit import _patch


class PatchEnvironFixture(tobiko.SharedFixture):

    original_environ = None
    patch_environ = None
    new_environ = None

    def __init__(self, **patch_environ):
        super(PatchEnvironFixture, self).__init__()
        self.patch_environ = patch_environ

    def setup_fixture(self):
        self.original_environ = os.environ
        self.new_environ = dict(os.environ, **self.patch_environ)
        os.environ = self.new_environ

    def cleanup_fixture(self):
        os.environ = self.original_environ


class FixtureManagerPatch(tobiko.FixtureManager, _patch.PatchFixture):

    def init_fixture(self, obj, name, fixture_id):
        fixture = super().init_fixture(obj=obj,
                                       name=name,
                                       fixture_id=fixture_id)
        self.addCleanup(tobiko.cleanup_fixture, fixture)
        return fixture

    def setup_fixture(self):
        self.patch(inspect.getmodule(tobiko.FixtureManager),
                   'FIXTURES',
                   self)


class TobikoUnitTest(_patch.PatchMixin, testtools.TestCase):

    patch_environ = {
        'http_proxy': 'http://127.0.0.1:88888',
        'https_proxy': 'http://127.0.0.1:88888',
        'no_proxy': '127.0.0.1'
    }

    def _get_test_method(self):
        method = super(TobikoUnitTest, self)._get_test_method()
        if inspect.iscoroutinefunction(method):

            @functools.wraps(method)
            def wrapped_test(*args, **kwargs):
                loop = asyncio.get_event_loop()
                task = loop.create_task(method(*args, **kwargs))
                loop.run_until_complete(task)

            return wrapped_test
        else:
            return method

    def setUp(self):
        super(TobikoUnitTest, self).setUp()
        # Protect from mis-configuring logging
        self.patch(log, 'setup')

        # Make sure each unit test uses it's own fixture manager
        self.fixture_manager = manager = FixtureManagerPatch()
        self.useFixture(manager)
        self.useFixture(PatchEnvironFixture(**self.patch_environ))

    def create_tempdir(self, *args, **kwargs):
        dir_path = tempfile.mkdtemp(*args, **kwargs)
        self.addCleanup(shutil.rmtree(dir_path, ignore_errors=True))
        return dir_path



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import subprocess
import sys

from oslo_log import log

import tobiko
from tobiko.shell.sh import _io
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _path
from tobiko.shell.sh import _process


LOG = log.getLogger(__name__)


def local_execute(command, environment=None, timeout: tobiko.Seconds = None,
                  shell=None, stdin=None, stdout=None, stderr=None,
                  expect_exit_status=0, **kwargs):
    """Execute command on local host using local shell"""
    process = local_process(command=command,
                            environment=environment,
                            timeout=timeout,
                            shell=shell,
                            stdin=stdin,
                            stdout=stdout,
                            stderr=stderr,
                            **kwargs)
    return _execute.execute_process(process=process,
                                    stdin=stdin,
                                    expect_exit_status=expect_exit_status)


def local_process(command, environment=None, current_dir=None,
                  timeout: tobiko.Seconds = None, shell=None, stdin=None,
                  stdout=None, stderr=True, sudo=None, network_namespace=None):
    return LocalShellProcessFixture(
        command=command, environment=environment, current_dir=current_dir,
        timeout=timeout, shell=shell, stdin=stdin, stdout=stdout,
        stderr=stderr, sudo=sudo, network_namespace=network_namespace)


class LocalExecutePathFixture(_path.ExecutePathFixture):

    executable_dirs = [os.path.realpath(os.path.join(d, 'bin'))
                       for d in [sys.prefix, sys.exec_prefix]]
    environ = dict(os.environ)


class LocalShellProcessFixture(_process.ShellProcessFixture):

    path_execute = tobiko.required_fixture(LocalExecutePathFixture)
    default_shell = True

    def create_process(self):
        tobiko.setup_fixture(self.path_execute)
        parameters = self.parameters
        args = self.command
        stdin = parameters.stdin and subprocess.PIPE or None
        stdout = parameters.stdout and subprocess.PIPE or None
        stderr = parameters.stderr and subprocess.PIPE or None
        env = merge_dictionaries(os.environ, parameters.environment)
        try:
            return subprocess.Popen(args=args,
                                    bufsize=parameters.buffer_size,
                                    shell=False,
                                    universal_newlines=False,
                                    env=env,
                                    cwd=parameters.current_dir,
                                    stdin=stdin,
                                    stdout=stdout,
                                    stderr=stderr)
        except FileNotFoundError as ex:
            LOG.debug(f"Error executing local command: args={args}")
            raise _exception.ShellCommandFailed(command=self.command,
                                                exit_status=-1,
                                                stdin=parameters.stdin,
                                                stdout=None,
                                                stderr=str(ex)) from ex

    def setup_stdin(self):
        self.stdin = _io.ShellStdin(delegate=self.process.stdin,
                                    buffer_size=self.parameters.buffer_size)

    def setup_stdout(self):
        self.stdout = _io.ShellStdout(delegate=self.process.stdout,
                                      buffer_size=self.parameters.buffer_size)

    def setup_stderr(self):
        self.stderr = _io.ShellStderr(delegate=self.process.stderr,
                                      buffer_size=self.parameters.buffer_size)

    def poll_exit_status(self):
        return self.process.poll()

    def _get_exit_status(self, timeout: tobiko.Seconds):
        try:
            exit_status = self.process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            LOG.exception("Failed waiting for subprocess termination")
            return None
        else:
            assert exit_status is not None
            return exit_status

    def kill(self):
        process = self.process
        LOG.debug('Killing local process: %r', self.command)
        try:
            process.kill()
        except Exception:
            LOG.exception('Failed killing local process: %r (PID=%r)',
                          self.command, self.pid)
        else:
            LOG.debug('Local process killed: %r (PID=%r)', self.command,
                      self.pid)

    @property
    def pid(self):
        process = self.process
        return process and process.pid or None


def merge_dictionaries(*dictionaries):
    merged = {}
    for d in dictionaries:
        if d:
            merged.update(d)
    return merged



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import asyncio
import inspect
import logging
import typing

from oslo_log import log

import tobiko
from tobiko.actor import _request


T = typing.TypeVar('T')


class ActorRef(tobiko.CallHandler):

    def __init__(self, actor_id: str, requests: _request.ActorRequestQueue):
        super().__init__()
        self.actor_id = actor_id
        self._requests = requests

    def send_request(self, method: str, **arguments):
        return self._requests.send_request(method=method, arguments=arguments)

    def _handle_call(self, method: typing.Callable, *args, **kwargs) -> None:
        arguments = inspect.signature(method).bind(
            None, *args, **kwargs).arguments
        arguments.pop('self', None)
        return self.send_request(method.__name__, **arguments)


def is_actor_method(obj):
    return getattr(obj, '__tobiko_actor_method__', False)


def actor_method(obj):
    if not callable(obj):
        raise TypeError(f"Actor method {obj} is not callable")

    if not inspect.iscoroutinefunction(obj):
        raise TypeError(f"Actor method {obj} is not async")

    name = getattr(obj, '__name__', None)
    if name is None or hasattr(ActorRef, name) or hasattr(Actor, name):
        raise TypeError(f"Invalid method name: '{name}'")

    obj.__tobiko_actor_method__ = True
    return obj


class Actor(tobiko.SharedFixture):
    max_queue_size: int = 0

    log: logging.LoggerAdapter
    loop: asyncio.AbstractEventLoop

    base_ref_class = ActorRef

    _actor_methods: typing.Dict[str, typing.Callable]
    _ref_class: type
    _ref: ActorRef
    _requests: _request.ActorRequestQueue
    _run_actor_task: asyncio.Task

    @property
    def actor_id(self) -> str:
        return self.fixture_name

    def setup_fixture(self):
        self.loop = self.get_loop()
        self.log = self.create_log()
        self._requests = self.create_request_queue()

        self._run_actor_task = self.loop.create_task(
            self._run_actor())

    @classmethod
    def ref_class(cls) -> type:
        try:
            return cls._ref_class
        except AttributeError:
            pass
        name = cls.__name__ + 'Ref'
        bases = cls.base_ref_class,
        namespace = {'__module__': cls.__module__,
                     'protocol_class': cls}
        return type(name, bases, namespace)

    @property
    def ref(self) -> ActorRef:
        try:
            return self._ref
        except AttributeError:
            pass
        ref_class = self.ref_class()
        self._ref = ref = ref_class(actor_id=self.actor_id,
                                    requests=self._requests)
        return ref

    def get_loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_event_loop()

    @classmethod
    def _get_actor_methods(cls) -> typing.Dict[str, typing.Callable]:
        try:
            return cls._actor_methods
        except AttributeError:
            pass
        cls._actor_methods = dict(inspect.getmembers(cls, is_actor_method))
        return cls._actor_methods

    def create_log(self):
        return log.getLogger(self.actor_id)

    def create_request_queue(self) -> _request.ActorRequestQueue:
        return _request.create_request_queue(max_size=self.max_queue_size,
                                             loop=self.loop)

    async def setup_actor(self):
        pass

    async def cleanup_actor(self):
        pass

    async def on_setup_error(self):
        self.log.exception("Actor setup error")

    async def on_request_error(
            self, request: typing.Optional[_request.ActorRequest]):
        self.log.exception(f"Actor request error: {request}")

    async def on_cleanup_error(self):
        self.log.exception("Actor cleanup error")

    async def _run_actor(self):
        try:
            await self.setup_actor()
        except Exception:
            await self.on_setup_error()
            await self._cleanup_actor()
        else:
            while True:
                request = None
                try:
                    request = await self._requests.receive_request()
                    if not isinstance(request, _request.ActorRequest):
                        raise TypeError(
                            f"Invalid actor request type: {request}")
                    await self._receive_request(request)
                except Exception:
                    await self.on_request_error(request=request)

    async def _cleanup_actor(self):
        try:
            await self.cleanup_actor()
        except Exception:
            await self.on_cleanup_error()

    def _get_actor_method(self, name: str) -> typing.Callable:
        methods = self._get_actor_methods()
        method = methods.get(name)
        if method is None:
            raise ValueError(f"Invalid request method name: {name}")
        return method

    async def _receive_request(self, request: _request.ActorRequest):
        method = self._get_actor_method(request.method)
        try:
            result = await method(self, **request.arguments)
        except Exception as ex:
            request.future.set_exception(ex)
        else:
            request.future.set_result(result)


def create_actor(obj: typing.Union[str, Actor, typing.Type[Actor]],
                 fixture_id: typing.Optional[str] = None,
                 manager=None) -> ActorRef:
    actor = tobiko.setup_fixture(obj,
                                 fixture_id=fixture_id,
                                 manager=manager)
    return actor.ref



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko


class PingException(tobiko.TobikoException):
    """Base ping command exception"""


class PingError(PingException):
    """Base ping error"""
    message = "{details!s}"


class LocalPingError(PingError):
    """Raised when local error happens"""


class SendToPingError(PingError):
    """Raised when sendto error happens"""


class ConnectPingError(PingError):
    """Raised when sendto error happens"""


class UnknowHostError(PingError):
    """Raised when unable to resolve host name"""


class BadAddressPingError(PingError):
    """Raised when passing wrong address to ping command"""
    message = "bad address: {address}"


class PingFailed(PingError, tobiko.FailureException):
    """Raised when ping timeout expires before reaching expected message count

    """
    message = ("timeout of {timeout} seconds expired after counting only "
               "{count} out of expected {expected_count} ICMP messages of "
               "type {message_type!r}")


class UnreachableHostsException(PingFailed):
    message = ("timeout of {timeout} seconds expired waiting for "
               "host(s) {hosts} to get pinged successfully")


class ReachableHostsException(PingFailed):
    message = ("timeout of {timeout} seconds expired waiting for "
               "hosts(s) {hosts} to fail getting pinged")


class UnsupportedPingOption(PingError):
    pass



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack.stacks import _nova


CONF = config.CONF


FEDORA_IMAGE_URL = (
    'https://download.fedoraproject.org/pub/fedora/linux/releases/34/Cloud/'
    'x86_64/images/Fedora-Cloud-Base-34-1.2.x86_64.qcow2')


class FedoraBaseImageFixture(glance.URLGlanceImageFixture):
    image_url = CONF.tobiko.fedora.image_url or FEDORA_IMAGE_URL
    image_name = CONF.tobiko.fedora.image_name
    image_file = CONF.tobiko.fedora.image_file
    disk_format = CONF.tobiko.fedora.disk_format or "qcow2"
    container_format = CONF.tobiko.fedora.container_format or "bare"
    username = CONF.tobiko.fedora.username or 'fedora'
    password = CONF.tobiko.fedora.password
    connection_timeout = CONF.tobiko.fedora.connection_timeout or 800.


class FedoraImageFixture(FedoraBaseImageFixture,
                         glance.CustomizedGlanceImageFixture):

    @property
    def run_commands(self) -> typing.List[str]:
        return super().run_commands + [
            'update-crypto-policies --set LEGACY']


class FedoraFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 256
    swap = 1024


class FedoraServerStackFixture(_nova.CloudInitServerStackFixture):

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_setup_fixture(FedoraImageFixture)

    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_setup_fixture(FedoraFlavorStackFixture)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

from tobiko.tripleo import containers


@containers.skip_unless_has_container_runtime()
class RuntimeRuntimeTest(testtools.TestCase):

    def test_get_container_runtime(self):
        runtime = containers.get_container_runtime()
        self.assertIsInstance(runtime, containers.ContainerRuntime)

    def test_list_containers(self):
        containers_list = containers.list_containers()
        for container in containers_list:
            self.assertIsNotNone(container)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

from tobiko.shell.iperf3 import _parameters
from tobiko.shell import sh


LOG = log.getLogger(__name__)


def get_iperf3_client_command(parameters: _parameters.Iperf3ClientParameters):
    interface = Iperf3Interface()
    return interface.get_iperf3_client_command(parameters)


class Iperf3Interface:

    def get_iperf3_client_command(
            self,
            parameters: _parameters.Iperf3ClientParameters) \
            -> sh.ShellCommand:
        options = self.get_iperf3_client_options(parameters=parameters)
        return sh.shell_command('iperf3') + options

    def get_iperf3_client_options(
            self,
            parameters: _parameters.Iperf3ClientParameters) \
            -> sh.ShellCommand:
        options = sh.ShellCommand(['-J'])
        options += self.get_client_mode_option(parameters.address)
        if parameters.port is not None:
            options += self.get_port_option(parameters.port)
        if parameters.timeout is not None:
            options += self.get_timeout_option(parameters.timeout)
        if parameters.bitrate is not None:
            options += self.get_bitrate_option(parameters.bitrate)
        if parameters.download is not None:
            options += self.get_download_option(parameters.download)
        if parameters.protocol is not None:
            options += self.get_protocol_option(parameters.protocol)
        return options

    @staticmethod
    def get_bitrate_option(bitrate: int):
        return ['-b', max(0, bitrate)]

    @staticmethod
    def get_client_mode_option(server_address: str):
        return ['-c', server_address]

    @staticmethod
    def get_download_option(download: bool):
        if download:
            return ['-R']
        else:
            return []

    @staticmethod
    def get_protocol_option(protocol: str):
        if protocol == 'tcp':
            return []
        elif protocol == 'udp':
            return ['-u']
        else:
            raise ValueError('iperf3 protocol values allowed: [tcp|udp]')

    @staticmethod
    def get_timeout_option(timeout: int):
        if timeout > 0:
            return ['-t', timeout]
        else:
            return []

    @staticmethod
    def get_port_option(port):
        return ['-p', port]



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect

from keystoneclient.v3 import endpoints
from octaviaclient.api.v2 import octavia as octaviaclient

from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.tests import unit
from tobiko.tests.unit import openstack
from tobiko.tests.unit.openstack import test_client


class KeystoneModulePatch(unit.PatchFixture):

    client = object()
    endpoint = endpoints.Endpoint(manager=None,
                                  info={'url': 'http://some/endpoint'})
    session = None
    name = None

    def setup_fixture(self):
        module = inspect.getmodule(octavia.OctaviaClientFixture)
        self.patch(module, 'keystone', self)

    def get_keystone_client(self, session):
        self.session = session
        return self.client

    def find_service_endpoint(self, name, client):
        self.name = name
        assert self.client is client
        return self.endpoint


class OctaviaClientFixtureTest(test_client.OpenstackClientFixtureTest):

    def setUp(self):
        super(OctaviaClientFixtureTest, self).setUp()
        self.useFixture(KeystoneModulePatch())

    def create_client(self, session=None):
        return octavia.OctaviaClientFixture(session=session)


class GetOctaviaClientTest(openstack.OpenstackTest):

    def setUp(self):
        super(GetOctaviaClientTest, self).setUp()
        self.useFixture(KeystoneModulePatch())

    def test_get_octavia_client(self, session=None, shared=True):
        client1 = octavia.get_octavia_client(session=session, shared=shared)
        client2 = octavia.get_octavia_client(session=session, shared=shared)
        if shared:
            self.assertIs(client1, client2)
        else:
            self.assertIsNot(client1, client2)
        self.assertIsInstance(client1, octaviaclient.OctaviaAPI)
        self.assertIsInstance(client2, octaviaclient.OctaviaAPI)

    def test_get_octavia_client_with_not_shared(self):
        self.test_get_octavia_client(shared=False)

    def test_get_octavia_client_with_session(self):
        session = keystone.get_keystone_session()
        self.test_get_octavia_client(session=session)


class OctaviaClientTest(openstack.OpenstackTest):

    def setUp(self):
        super(OctaviaClientTest, self).setUp()
        self.useFixture(KeystoneModulePatch())

    def test_octavia_client_with_none(self):
        default_client = octavia.get_octavia_client()
        client = octavia.octavia_client(None)
        self.assertIsInstance(client, octaviaclient.OctaviaAPI)
        self.assertIs(default_client, client)

    def test_octavia_client_with_client(self):
        default_client = octavia.get_octavia_client()
        client = octavia.octavia_client(default_client)
        self.assertIsInstance(client, octaviaclient.OctaviaAPI)
        self.assertIs(default_client, client)

    def test_octavia_client_with_fixture(self):
        fixture = octavia.OctaviaClientFixture()
        client = octavia.octavia_client(fixture)
        self.assertIsInstance(client, octaviaclient.OctaviaAPI)
        self.assertIs(client, fixture.client)



# Copyright 2020 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import mock

import tobiko
from tobiko.tests import unit


class TimeTest(unit.TobikoUnitTest):

    def test_time(self):
        mock_time = self.patch_time()
        initial_time = mock_time.current_time

        self.assertEqual(mock_time.current_time, tobiko.time())
        self.assertEqual(mock_time.current_time, tobiko.time())
        self.assertEqual(mock_time.current_time, tobiko.time())

        final_time = mock_time.current_time
        self.assertEqual(3 * mock_time.time_increment,
                         final_time - initial_time)
        mock_time.time.assert_has_calls([mock.call()] * 3, any_order=True)

    def test_sleep_with_float(self):
        self._test_sleep(20.)

    def test_sleep_with_int(self):
        self._test_sleep(10)

    def test_sleep_with_str(self):
        self._test_sleep('5.')

    def test_sleep_with_bytes(self):
        self._test_sleep(b"15")

    def test_sleep_with_zero(self):
        self._test_sleep(0.)

    def test_sleep_with_inf(self):
        self._test_sleep(float('inf'))

    def test_sleep_with_negative(self):
        self._test_sleep(-1.)

    def test_sleep_with_none(self):
        self._test_sleep(None)

    def _test_sleep(self, seconds: tobiko.Seconds):
        mock_time = self.patch_time()
        initial_time = mock_time.current_time

        tobiko.sleep(seconds)

        expected = seconds and max(0., float(seconds)) or 0.
        final_time = mock_time.current_time
        mock_time.sleep.assert_called_once_with(expected)
        self.assertEqual(expected, final_time - initial_time)

    def test_to_seconds_with_float(self):
        self._test_to_seconds(20.)

    def test_to_seconds_with_int(self):
        self._test_to_seconds(20)

    def test_to_seconds_with_str(self):
        self._test_to_seconds("5")

    def test_to_seconds_with_bytes(self):
        self._test_to_seconds(b"15")

    def test_to_seconds_with_zero(self):
        self._test_to_seconds(0.)

    def test_to_seconds_with_inf(self):
        self._test_to_seconds(float('inf'))

    def test_to_seconds_with_negative(self):
        self._test_to_seconds(-1.)

    def test_to_seconds_with_none(self):
        self._test_to_seconds(None)

    def _test_to_seconds(self, seconds: tobiko.Seconds):
        result = tobiko.to_seconds(seconds)
        if seconds is None:
            self.assertIsNone(result)
        else:
            self.assertEqual(max(0., float(seconds)), result)

    def test_to_seconds_float_with_float(self):
        self._test_to_seconds_float(20.)

    def test_to_seconds_float_with_int(self):
        self._test_to_seconds_float(20)

    def test_to_seconds_float_with_str(self):
        self._test_to_seconds_float("5")

    def test_to_seconds_float_with_bytes(self):
        self._test_to_seconds_float(b"15")

    def test_to_seconds_float_with_zero(self):
        self._test_to_seconds_float(0.)

    def test_to_seconds_float_with_inf(self):
        self._test_to_seconds_float(float('inf'))

    def test_to_seconds_float_with_negative(self):
        self._test_to_seconds_float(-1.)

    def test_to_seconds_float_with_none(self):
        self._test_to_seconds_float(None)

    def _test_to_seconds_float(self, seconds: tobiko.Seconds):
        result = tobiko.to_seconds_float(seconds)
        expected = seconds and max(0., float(seconds)) or 0.
        self.assertEqual(expected, result)

    def test_true_seconds(self):
        self.assertEqual([], tobiko.true_seconds())
        self.assertEqual([], tobiko.true_seconds(None))
        self.assertEqual([], tobiko.true_seconds(None, None))
        self.assertEqual([1., 2., 3., 4.],
                         tobiko.true_seconds(None, 1, None, 2., '3', None, 4,
                                             None))

    def test_min_seconds(self):
        self.assertIsNone(tobiko.min_seconds())
        self.assertIsNone(tobiko.min_seconds(None))
        self.assertIsNone(tobiko.min_seconds(None, None))
        self.assertEqual(1.5, tobiko.min_seconds(3, 2., '1.5'))

    def test_max_seconds(self):
        self.assertIsNone(tobiko.max_seconds())
        self.assertIsNone(tobiko.max_seconds(None))
        self.assertIsNone(tobiko.max_seconds(None, None))
        self.assertEqual(3., tobiko.max_seconds(2., 3, '1.5'))



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import collections
import sys
import typing  # noqa

from oslo_log import log
import testtools


LOG = log.getLogger(__name__)


class TobikoException(Exception):
    """Base Tobiko Exception.

    To use this class, inherit from it and define attribute 'message' string.
    If **properties parameters is given, then it will format message string
    using properties as key-word arguments.

    Example:

        class MyException(TobikoException):
            message = "This exception occurred because of {reason}"

        try:
            raise MyException(reason="something went wrong")
        except MyException as ex:

            # It should print:
            #   This exception occurred because of something went wrong
            print(ex)

            # It should print:
            #   something went wrong
            print(ex.reason)

    :attribute message: the message to be printed out.
    """

    message = "unknown reason"

    def __init__(self, message=None, **properties):
        # pylint: disable=exception-message-attribute
        message = message or self.message
        if properties:
            message = message.format(**properties)
        self.message = message
        self._properties = properties or {}
        super(TobikoException, self).__init__(message)

    def __getattr__(self, name):
        try:
            return self._properties[name]
        except KeyError as ex:
            raise AttributeError(f"{self!r} object has no attribute "
                                 f"'{name}'") from ex

    def __repr__(self):
        return "{class_name}({message!r})".format(
            class_name=type(self).__name__,
            message=self.message)

    def __eq__(self, other):
        return type(self) == type(other) and str(self) == str(other)

    def __hash__(self):
        return hash(type(self)) + hash(str(self))


def check_valid_type(obj, *valid_types):
    if not isinstance(obj, valid_types):
        types_str = ", ".join(str(t) for t in valid_types)
        message = f"Object {obj!r} is not of a valid type ({types_str})"
        raise TypeError(message)
    return obj


class ExceptionInfo(collections.namedtuple('ExceptionInfo',
                                           ['type', 'value', 'traceback'])):

    reraise_on_exit = True

    def __enter__(self):
        return self

    def __bool__(self):
        return self.type is not None

    def __exit__(self, _type, _value, _traceback):
        if self.reraise_on_exit:
            self.reraise()

    def reraise(self):
        if self.type is not None:
            reraise(*self)


def reraise(tp, value, tb=None):
    try:
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value
    finally:
        value = None
        tb = None


def exc_info(reraise=True):
    # pylint: disable=redefined-outer-name
    info = ExceptionInfo(*sys.exc_info())
    info.reraise_on_exit = reraise
    return info


def log_unhandled_exception(exc_type, exc_value, ex_tb):
    LOG.exception("Unhandled exception",
                  exc_info=(exc_type, exc_value, ex_tb))


@contextlib.contextmanager
def handle_multiple_exceptions(handle_exception=log_unhandled_exception):
    exception: typing.Optional[typing.Tuple] = None
    try:
        yield
    except testtools.MultipleExceptions as ex:
        exc_infos = list_exc_infos()
        if exc_infos:
            exception = exc_infos[0]
            for info in exc_infos[1:]:
                try:
                    handle_exception(*info)
                except Exception:
                    LOG.exception("Error handling multiple exceptions")
        else:
            LOG.debug(f"Empty MultipleExceptions: '{ex}'")

    if exception is not None:
        reraise(*exception)


def list_exc_infos(exc_info=None):
    # pylint: disable=redefined-outer-name
    exc_info = exc_info or sys.exc_info()
    result = []
    if exc_info[0]:
        visited = set()
        visiting = [exc_info]
        while visiting:
            exc_info = visiting.pop()
            _, exc, _ = exc_info
            if exc not in visited:
                visited.add(exc)
                if isinstance(exc, testtools.MultipleExceptions):
                    visiting.extend(reversed(exc.args))
                else:
                    result.append(exc_info)
    return result



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import io
import os
from urllib import parse

import netaddr
from oslo_log import log
import paramiko

import tobiko


LOG = log.getLogger(__name__)


def ssh_config(config_files=None):
    if config_files:
        fixture = SSHConfigFixture(config_files=config_files)
    else:
        fixture = SSHConfigFixture
    return tobiko.setup_fixture(fixture)


def ssh_host_config(host: str,
                    config_files=None):
    return ssh_config(config_files=config_files).lookup(host)


class SSHDefaultConfigFixture(tobiko.SharedFixture):

    conf = None

    def setup_fixture(self):
        self.conf = tobiko.tobiko_config().ssh

    def __getattr__(self, name):
        return getattr(self.conf, name)


def get_ssh_host_url(host: str) -> parse.ParseResult:
    try:
        ip_address = netaddr.IPAddress(host)
    except netaddr.AddrFormatError:
        pass
    else:
        if ip_address.version == 6:
            host = f"[{host}]"
    return parse.urlparse(f"ssh://{host}")


class SSHConfigFixture(tobiko.SharedFixture):

    default = tobiko.required_setup_fixture(SSHDefaultConfigFixture)

    config_files = None
    config = None

    def __init__(self, config_files=None):
        super(SSHConfigFixture, self).__init__()
        if config_files:
            self.config_files = tuple(config_files)
        else:
            self.config_files = self.default.config_files

    def setup_fixture(self):
        self.setup_ssh_config()

    def setup_ssh_config(self):
        self.config = paramiko.SSHConfig()
        for config_file in self.config_files:
            config_file = tobiko.tobiko_config_path(config_file)
            if os.path.exists(config_file):
                LOG.debug("Parsing %r config file...", config_file)
                try:
                    with io.open(config_file, 'rt',
                                 encoding="utf-8") as f:
                        self.config.parse(f)
                except Exception as ex:
                    LOG.exception(f"Error parsing '{config_file}' SSH config "
                                  f"file: {ex}")
                else:
                    LOG.debug("File %r parsed.", config_file)

    def lookup(self,
               host: str = None,
               hostname: str = None,
               username: str = None,
               port: int = None):

        if host and ('@' in host or ':' in host):
            host_url = get_ssh_host_url(host)
            hostname = hostname or host_url.hostname or None
            username = username or host_url.username or None
            port = port or host_url.port or None
        else:
            hostname = hostname or host

        if hostname and self.config:
            host_config: dict = self.config.lookup(hostname)
        else:
            host_config = {}

        # remove unsupported directive
        include_files = host_config.pop('include', None)
        if include_files:
            LOG.warning('Ignoring unsupported directive: Include %s',
                        include_files)
        if hostname:
            host_config.setdefault('hostname', hostname)
        if username:
            host_config.setdefault('user', username)
        if port:
            host_config.setdefault('port', port)
        return SSHHostConfig(host=host,
                             ssh_config=self,
                             host_config=host_config,
                             config_files=self.config_files)

    def __repr__(self):
        return "{class_name!s}(config_files={config_files!r})".format(
            class_name=type(self).__name__, config_files=self.config_files)


class SSHHostConfig(collections.namedtuple('SSHHostConfig', ['host',
                                                             'ssh_config',
                                                             'host_config',
                                                             'config_files'])):

    default = tobiko.required_setup_fixture(SSHDefaultConfigFixture)

    @property
    def hostname(self):
        return self.host_config.get('hostname', self.host)

    @property
    def port(self):
        return (self.host_config.get('port') or
                self.default.port)

    @property
    def username(self):
        return (self.host_config.get('user') or
                self.default.username)

    @property
    def key_filename(self):
        key_filename = []
        host_config_key_files = self.host_config.get('identityfile')
        if host_config_key_files:
            for filename in host_config_key_files:
                if filename:
                    key_filename.append(tobiko.tobiko_config_path(filename))

        default_key_files = self.default.key_file
        if default_key_files:
            for filename in default_key_files:
                if filename:
                    key_filename.append(tobiko.tobiko_config_path(filename))

        return key_filename

    @property
    def proxy_jump(self):
        proxy_jump = (self.host_config.get('proxyjump') or
                      self.default.proxy_jump)
        if not proxy_jump:
            return None

        proxy_hostname = self.ssh_config.lookup(proxy_jump).hostname
        if ({proxy_jump, proxy_hostname} & {self.host, self.hostname}):
            # Avoid recursive proxy jump definition
            return None

        return proxy_jump

    @property
    def proxy_command(self):
        return (self.host_config.get('proxycommand') or
                self.default.proxy_command)

    @property
    def allow_agent(self):
        return (is_yes(self.host_config.get('forwardagent')) or
                self.default.allow_agent)

    @property
    def compress(self):
        return (is_yes(self.host_config.get('compression')) or
                self.default.compress)

    @property
    def timeout(self):
        return (self.host_config.get('connetcttimeout') or
                self.default.timeout)

    @property
    def connection_attempts(self):
        return (self.host_config.get('connectionattempts') or
                self.default.connection_attempts)

    @property
    def connection_interval(self):
        return self.default.connection_interval

    @property
    def connection_timeout(self):
        return self.default.connection_timeout

    @property
    def connect_parameters(self):
        return dict(hostname=self.hostname,
                    port=self.port,
                    username=self.username,
                    key_filename=self.key_filename,
                    compress=self.compress,
                    timeout=self.timeout,
                    allow_agent=self.allow_agent,
                    connection_attempts=self.connection_attempts,
                    connection_interval=self.connection_interval,
                    connection_timeout=self.connection_timeout)


def is_yes(value):
    return value and str(value).lower() == 'yes'
